#!/usr/bin/env python
# crate_anon/anonymise/models.py

"""
To create a SQLAlchemy Table programmatically:
    http://docs.sqlalchemy.org/en/latest/core/schema.html
    http://stackoverflow.com/questions/5424942/sqlalchemy-model-definition-at-execution  # noqa
    http://stackoverflow.com/questions/2580497/database-on-the-fly-with-scripting-languages/2580543#2580543  # noqa

To create a SQLAlchemy ORM programmatically:
    http://stackoverflow.com/questions/2574105/sqlalchemy-dynamic-mapping/2575016#2575016  # noqa
"""

import logging
import random

from sqlalchemy import (
    Column,
    MetaData,
    Text,
)
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session

from crate_anon.anonymise.config_singleton import config
from crate_anon.anonymise.constants import (
    MAX_TRID,
    MYSQL_TABLE_KWARGS,
    PidType,
    TridType,
)
from crate_anon.anonymise.scrub import PersonalizedScrubber
from crate_anon.common.sqla import exists_orm

log = logging.getLogger(__name__)
admin_meta = MetaData()
AdminBase = declarative_base(metadata=admin_meta)


class PatientInfo(AdminBase):
    __tablename__ = 'secret_map'
    __table_args__ = MYSQL_TABLE_KWARGS

    pid = Column(
        'pid', PidType,
        primary_key=True, autoincrement=False,
        doc="Patient ID (PID) (PK)")
    rid = Column(
        'rid', config.SqlTypeEncryptedPid,
        nullable=False, unique=True,
        doc="Research ID (RID)")
    trid = Column(
        'trid', TridType,
        unique=True,
        doc="Transient integer research ID (TRID)")
    mpid = Column(
        'mpid', PidType,
        doc="Master patient ID (MPID)")
    mrid = Column(
        'mrid', config.SqlTypeEncryptedPid,
        doc="Master research ID (MRID)")
    scrubber_hash = Column(
        'scrubber_hash', config.SqlTypeEncryptedPid,
        doc="Scrubber hash (for change detection)")
    patient_scrubber_text = Column(
        "_raw_scrubber_patient", Text,
        doc="Raw patient scrubber (for debugging only)")
    tp_scrubber_text = Column(
        "_raw_scrubber_tp", Text,
        doc="Raw third-party scrubber (for debugging only)")

    def ensure_rid(self) -> None:
        assert self.pid is not None
        if self.rid is not None:
            return
        self.rid = config.encrypt_primary_pid(self.pid)

    def ensure_trid(self, session: Session) -> None:
        assert self.pid is not None
        if self.trid is not None:
            return
        # noinspection PyTypeChecker
        self.trid = TridRecord.get_trid(session, self.pid)

    def set_mpid(self, mpid: int) -> None:
        self.mpid = mpid
        self.mrid = config.encrypt_master_pid(self.mpid)

    def set_scrubber_info(self, scrubber: PersonalizedScrubber) -> None:
        self.scrubber_hash = scrubber.get_hash()
        if config.save_scrubbers:
            self.patient_scrubber_text = scrubber.get_patient_regex_string()
            self.tp_scrubber_text = scrubber.get_tp_regex_string()
        else:
            self.patient_scrubber_text = None
            self.tp_scrubber_text = None


class TridRecord(AdminBase):
    __tablename__ = 'secret_trid_cache'
    __table_args__ = MYSQL_TABLE_KWARGS

    pid = Column(
        "pid", PidType,
        primary_key=True, autoincrement=False,
        doc="Patient ID (PID) (PK)")
    trid = Column(
        "trid", TridType,
        nullable=False, unique=True,
        doc="Transient integer research ID (TRID)")

    @classmethod
    def get_trid(cls, session: Session, pid: int) -> int:
        try:
            obj = session.query(cls).filter(cls.pid == pid).one()
            return obj.trid
        except NoResultFound:
            return cls.new_trid(session, pid)

    @classmethod
    def new_trid(cls, session: Session, pid: int) -> int:
        """
        We check for existence by inserting and asking the database if it's
        happy, not by asking the database if it exists (since other processes
        may be doing the same thing at the same time).
        """
        while True:
            session.begin_nested()
            candidate = random.randint(1, MAX_TRID)
            log.debug("Trying candidate TRID: {}".format(candidate))
            obj = cls(pid=pid, trid=candidate)
            try:
                session.add(obj)
                session.commit()  # may raise IntegrityError
                return candidate
            except IntegrityError:
                session.rollback()


class OptOutPid(AdminBase):
    __tablename__ = 'opt_out_pid'
    __table_args__ = MYSQL_TABLE_KWARGS

    pid = Column(
        'pid', PidType,
        primary_key=True,
        doc="Patient ID")

    @classmethod
    def opting_out(cls, session: Session, pid: int) -> bool:
        return exists_orm(session, cls, cls.pid == pid)

    @classmethod
    def add(cls, session: Session, pid: int) -> None:
        log.debug("Adding opt-out for PID {}".format(pid))
        newthing = cls(pid=pid)
        session.merge(newthing)
        # http://stackoverflow.com/questions/12297156/fastest-way-to-insert-object-if-it-doesnt-exist-with-sqlalchemy  # noqa


class OptOutMpid(AdminBase):
    __tablename__ = 'opt_out_mpid'
    __table_args__ = MYSQL_TABLE_KWARGS

    mpid = Column(
        'mpid', PidType,
        primary_key=True,
        doc="Patient ID")

    @classmethod
    def opting_out(cls, session: Session, mpid: int) -> bool:
        return exists_orm(session, cls, cls.mpid == mpid)

    @classmethod
    def add(cls, session: Session, mpid: int) -> None:
        log.debug("Adding opt-out for MPID {}".format(mpid))
        newthing = cls(mpid=mpid)
        session.merge(newthing)



#!/usr/bin/env python
# crate_anon/crateweb/config/constants.py

CRATEWEB_CONFIG_ENV_VAR = 'CRATE_WEB_LOCAL_SETTINGS'
CHERRYPY_EXTRA_ARGS_ENV_VAR = 'CRATE_CHERRYPY_ARGS'
CELERY_APP_NAME = 'crate_anon.crateweb.consent'



#!/usr/bin/env python
# crate_anon/tools/launch_flower.py

import subprocess

from crate_anon.crateweb.config.constants import CELERY_APP_NAME


def main():
    cmdargs = [
        "celery",
        "-A", CELERY_APP_NAME,
        "flower"
    ]
    print("Launching Flower: {}".format(cmdargs))
    subprocess.call(cmdargs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/crateweb/extra/middleware.py

import logging
from re import compile
import sys

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
# from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.views.debug import technical_500_response
from django.utils.cache import add_never_cache_headers

log = logging.getLogger(__name__)


# =============================================================================
# UserBasedExceptionMiddleware
# =============================================================================
# From p436 of Greenfield & Greenfield (2015), "Two Scoops of Django:
# Best Practices for Django 1.8".

class UserBasedExceptionMiddleware(object):
    # noinspection PyUnusedLocal
    @staticmethod
    def process_exception(request, exception):
        if request.user.is_superuser:
            return technical_500_response(request, *sys.exc_info())


# =============================================================================
# LoginRequiredMiddleware
# =============================================================================

# -----------------------------------------------------------------------------
# 1.
# -----------------------------------------------------------------------------
"""
Middleware to require login for all pages.

Copyright (C) 2008 Ryan Witt
Licensed under the Creative Commons Attribution 3.0 United States License
http://onecreativeblog.com/post/59051248/django-login-required-middleware

Modified according to: https://djangosnippets.org/snippets/2845/
"""

# EXEMPT_URLS = [compile(settings.LOGIN_URL.lstrip('/'))]
# if hasattr(settings, 'LOGIN_EXEMPT_URLS'):
#     EXEMPT_URLS += [compile(expr) for expr in settings.LOGIN_EXEMPT_URLS]
#
#
# class LoginRequiredMiddleware:
#     """
#     Middleware that requires a user to be authenticated to view any page
#     other than LOGIN_URL. Exemptions to this requirement can optionally be
#     specified in settings via a list of regular expressions in
#     LOGIN_EXEMPT_URLS (which you can copy from your urls.py).
#
#     Requires authentication middleware and template context processors to be
#     loaded. You'll get an error if they aren't.
#     """
#     def process_request(self, request):
#         assert hasattr(request, 'user'), "The Login Required middleware\
#  requires authentication middleware to be installed. Edit your\
#  MIDDLEWARE_CLASSES setting to insert\
#  'django.contrib.auth.middleware.AuthenticationMiddleware'. If that doesn't\
#  work, ensure your TEMPLATE_CONTEXT_PROCESSORS setting includes\
#  'django.core.context_processors.auth'."
#         if not request.user.is_authenticated():
#             path = request.path_info.lstrip('/')
#             if not any(m.match(path) for m in EXEMPT_URLS):
#                 # (1) Simple version:
#                 # return HttpResponseRedirect(settings.LOGIN_URL)
#                 # (2) With 'next' parameter:
#                 path = request.get_full_path()
#                 return redirect_to_login(path, settings.LOGIN_URL,
#                                          REDIRECT_FIELD_NAME)


# -----------------------------------------------------------------------------
# 2. Alternative!
# -----------------------------------------------------------------------------
# http://stackoverflow.com/questions/3214589/

# class LoginRequiredMiddleware(object):
#     """
#     For an exempt view:
#         def someview(request, *args, **kwargs):
#             # body of view
#         someview.login_required = False
#
#         class SomeView(View):
#             login_required = False
#             # body of view
#     """
#     def process_view(self, request, view_func, view_args, view_kwargs):
#         if not getattr(view_func, 'login_required', True):
#             return None  # exempt
#         return login_required(view_func)(request, *view_args, **view_kwargs)


# -----------------------------------------------------------------------------
# 3. RNC; composite of those patterns.
# -----------------------------------------------------------------------------

EXEMPT_URLS = [compile(settings.LOGIN_URL.lstrip('/'))]
if hasattr(settings, 'LOGIN_EXEMPT_URLS'):
    EXEMPT_URLS += [compile(expr.lstrip('/'))
                    for expr in settings.LOGIN_EXEMPT_URLS]


# noinspection PyClassHasNoInit
class LoginRequiredMiddleware:
    """
    Middleware that requires a user to be authenticated to view any page other
    than LOGIN_URL. Exemptions to this requirement can optionally be specified
    in settings via a list of regular expressions in LOGIN_EXEMPT_URLS (which
    you can copy from your urls.py).

    Requires authentication middleware and template context processors to be
    loaded. You'll get an error if they aren't.

    Other way of doing exemptions, for an exempt view:

        def someview(request, *args, **kwargs):
            # body of view
        someview.login_required = False

        class SomeView(View):
            login_required = False
            # body of view
    """

    # noinspection PyUnusedLocal
    @staticmethod
    def process_view(request, view_func, view_args, view_kwargs):
        # log.critical("LoginRequiredMiddleware.process_view")
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "The Login Required middleware requires authentication "
                "middleware to be installed. Edit your MIDDLEWARE_CLASSES "
                "setting to insert "
                "'django.contrib.auth.middleware.AuthenticationMiddleware'. "
                "If that doesn't work, ensure your "
                "TEMPLATE_CONTEXT_PROCESSORS setting includes "
                "'django.core.context_processors.auth'.")
        if request.user.is_authenticated():
            # log.critical("is_authenticated")
            return None  # OK
        if not getattr(view_func, 'login_required', True):
            # log.critical("function exempt from login_requred")
            return None  # OK, exempt
        path = request.path_info.lstrip('/')
        # Path might look like 'login/' regardless of Django mount point
        if any(m.match(path) for m in EXEMPT_URLS):
            # log.critical("URL exempt from login_requred")
            return None  # OK, exempt
        fullpath = request.get_full_path()
        return redirect_to_login(fullpath, reverse(settings.LOGIN_VIEW_NAME),
                                 REDIRECT_FIELD_NAME)


# =============================================================================
# DisableClientSideCachingMiddleware
# =============================================================================
# http://stackoverflow.com/questions/2095520/fighting-client-side-caching-in-django  # noqa

class DisableClientSideCachingMiddleware(object):
    # noinspection PyUnusedLocal
    @staticmethod
    def process_response(request, response):
        add_never_cache_headers(response)
        return response



#!/usr/bin/env python
# crate_anon/anonymise/crateconfigparser.py

"""
Slightly extended ConfigParser.
"""

import ast
import configparser
from typing import Dict, Iterable, Iterator, Generic, List, Optional
# http://mypy-lang.org/examples.html
# https://www.python.org/dev/peps/pep-0484/
# https://docs.python.org/3/library/typing.html

from crate_anon.anonymise.dbholder import DatabaseHolder


def gen_lines(multiline: str) -> Iterator[str]:
    for line in multiline.splitlines():
        line = line.strip()
        if line:
            yield line


def gen_words(lines: Iterable[str]) -> Iterator[str]:
    for line in lines:
        for word in line.split():
            yield word


def gen_ints(words: Iterable[str],
             minimum: int = None,
             maximum: int = None,
             suppress_errors: bool = False) -> Iterator[int]:
    for word in words:
        try:
            value = int(word)
            if minimum is not None:
                if value < minimum:
                    raise ValueError("Value {} less than minimum of {}".format(
                        value, minimum))
            if maximum is not None:
                if value > maximum:
                    raise ValueError("Value {} more than maximum of {}".format(
                        value, maximum))
            yield value
        except ValueError:
            if not suppress_errors:
                raise


DB_SAFE_CONFIG_FWD_REF = "DatabaseSafeConfig"


class ExtendedConfigParser(configparser.ConfigParser):
    def __init__(self, *args, **kwargs) -> None:
        kwargs['interpolation'] = None
        kwargs['inline_comment_prefixes'] = ('#', ';')
        # 'converters': Python 3.5 and up
        super().__init__(*args, **kwargs)

    # Use the underlying ConfigParser class for e.g.
    #       getboolean(section, option)

    # UNNECESSARY: USE inline_comment_prefixes
    #
    # @staticmethod
    # def strip_inline_comment(text, comment_chars=None):
    #     if comment_chars is None:
    #         comment_chars = ['#', ';']  # standard for ConfigParser
    #     absent = -1
    #     commentpos = absent
    #     for cc in comment_chars:
    #         pos = text.find(cc)
    #         if pos != absent:
    #             commentpos = min(commentpos,
    #                              pos) if commentpos != absent else pos
    #     if commentpos == absent:
    #         return text.strip()
    #     return text[:commentpos].strip()

    @staticmethod
    def raise_missing(section: str,
                      option: str) -> None:
        raise ValueError("Config section {}: missing parameter: {}".format(
            section, option))

    def get_str(self,
                section: str,
                option: str,
                required: bool = False,
                default: str = None) -> Optional[str]:
        if required and default is not None:
            raise AssertionError("required and default are incompatible")
        s = self.get(section, option, fallback=default)
        if required and not s:
            self.raise_missing(section, option)
        return s

    def get_str_list(self,
                     section: str,
                     option: str,
                     as_words: bool = True,
                     lower: bool = False,
                     required: bool = False) -> List[str]:
        multiline = self.get(section, option, fallback='')
        if lower:
            multiline = multiline.lower()
        if as_words:
            result = list(gen_words(gen_lines(multiline)))
        else:  # as lines
            result = list(gen_lines(multiline))
        if required and not result:
            self.raise_missing(section, option)
        return result

    def get_int_default_if_failure(self,
                                   section: str,
                                   option: str,
                                   default: int = None) -> Optional[int]:
        try:
            return self.getint(section, option, fallback=default)
        except ValueError:  # e.g. invalid literal for int() with base 10
            return default

    def get_int_list(self,
                     section: str,
                     option: str,
                     minimum: int = None,
                     maximum: int = None,
                     suppress_errors: bool = True) -> List[int]:
        multiline = self.get(section, option, fallback='')
        return list(gen_ints(gen_words(gen_lines(multiline)),
                             minimum=minimum,
                             maximum=maximum,
                             suppress_errors=suppress_errors))

    def get_pyvalue_list(self,
                         section: str,
                         option: str,
                         default: Generic = None) -> List[Generic]:
        default = default or []
        strvalue = self.get(section, option, fallback=None)
        if not strvalue:
            return default
        pyvalue = ast.literal_eval(strvalue)
        # Now, make sure it's a list:
        # http://stackoverflow.com/questions/1835018
        if not isinstance(pyvalue, list):
            raise ValueError("Option {} must evaluate to a Python list "
                             "using ast.literal_eval()".format(option))
        return pyvalue

    def get_database(self,
                     section: str,
                     dbname: str = None,
                     srccfg: DB_SAFE_CONFIG_FWD_REF = None,
                     with_session: bool = False,
                     with_conn: bool = False,
                     reflect: bool = False) -> DatabaseHolder:
        dbname = dbname or section
        url = self.get_str(section, 'url', required=True)
        return DatabaseHolder(dbname, url, srccfg=srccfg,
                              with_session=with_session,
                              with_conn=with_conn,
                              reflect=reflect)

    def get_env_dict(self,
                     section: str,
                     parent_env: Optional[Dict] = None) -> Dict:
        if parent_env:
            env = parent_env.copy()
        else:
            env = {}
        newitems = {(str(k), str(v))
                    for k, v in self.items(section)}
        # items() returns a list of (name, value) tuples
        env.update(newitems)
        return env



#!/usr/bin/env python
# crate_anon/crateweb/consent/management/commands/test_email.py

from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Test email to RDBM (without saving to database)"

    def handle(self, *args, **options):
        sender = settings.EMAIL_SENDER
        recipient = settings.DEVELOPER_EMAIL
        send_mail(
            'CRATE test e-mail',
            'Message body',
            sender,
            [recipient],
            fail_silently=False,
        )
        self.stdout.write("Successfully sent e-mail from {} to {}".format(
            sender, recipient))



#!/usr/bin/env python
# crate_anon/anonymise/altermethod.py

import logging

# don't import config: circular dependency would have to be sorted out
from crate_anon.anonymise.constants import ALTERMETHOD

log = logging.getLogger(__name__)


# =============================================================================
# AlterMethod
# =============================================================================

class AlterMethod(object):
    def __init__(self,
                 text_value: str = None,
                 scrub: bool = False,
                 truncate_date: bool = False,
                 extract_from_filename: bool = False,
                 extract_from_blob: bool = False,
                 skip_if_text_extract_fails: bool = False,
                 extract_ext_field: str = "",
                 # html_escape: bool = False,
                 html_unescape: bool = False,
                 html_untag: bool = False) -> None:
        self.scrub = scrub
        self.truncate_date = truncate_date
        self.extract_text = (extract_from_filename or extract_from_blob)
        self.extract_from_blob = extract_from_blob
        self.extract_from_filename = extract_from_filename
        self.skip_if_text_extract_fails = skip_if_text_extract_fails
        self.extract_ext_field = extract_ext_field
        # self.html_escape = html_escape
        self.html_unescape = html_unescape
        self.html_untag = html_untag
        if text_value is not None:
            self.set_from_text(text_value)

    def set_from_text(self, value: str) -> None:
        """
        Convert the alter_method field (from the data dictionary) to a bunch of
        boolean/simple fields.
        """
        self.scrub = False
        self.truncate_date = False
        self.extract_text = False
        self.extract_from_blob = False
        self.extract_from_filename = False
        self.skip_if_text_extract_fails = False
        self.extract_ext_field = ""
        if value == ALTERMETHOD.TRUNCATEDATE.value:
            self.truncate_date = True
        elif value == ALTERMETHOD.SCRUBIN.value:
            self.scrub = True
        elif value.startswith(ALTERMETHOD.BIN2TEXT.value):
            if "=" not in value:
                raise ValueError(
                    "Bad format for alter method: {}".format(value))
            secondhalf = value[value.index("=") + 1:]
            if not secondhalf:
                raise ValueError(
                    "Missing filename/extension field in alter method: "
                    "{}".format(value))
            self.extract_text = True
            self.extract_from_blob = True
            self.extract_ext_field = secondhalf
        elif value == ALTERMETHOD.FILENAME2TEXT.value:
            self.extract_text = True
            self.extract_from_filename = True
        elif value == ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS.value:
            self.skip_if_text_extract_fails = True
        # elif value == ALTERMETHOD.HTML_ESCAPE:
        #     self.html_escape = True
        elif value == ALTERMETHOD.HTML_UNESCAPE.value:
            self.html_unescape = True
        elif value == ALTERMETHOD.HTML_UNTAG.value:
            self.html_untag = True
        else:
            raise ValueError("Bad alter_method part: {}".format(value))

    def get_text(self) -> str:
        """
        Return the alter_method fragment from the working fields.
        """
        if self.truncate_date:
            return ALTERMETHOD.TRUNCATEDATE.value
        if self.scrub:
            return ALTERMETHOD.SCRUBIN.value
        if self.extract_text:
            if self.extract_from_blob:
                return (ALTERMETHOD.BIN2TEXT.value + "=" +
                        self.extract_ext_field)
            else:
                return ALTERMETHOD.FILENAME2TEXT.value
        if self.skip_if_text_extract_fails:
            return ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS.value
        # if self.html_escape:
        #     return ALTERMETHOD.HTML_ESCAPE.value
        if self.html_unescape:
            return ALTERMETHOD.HTML_UNESCAPE.value
        if self.html_untag:
            return ALTERMETHOD.HTML_UNTAG.value
        return ""



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-28 13:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('userprofile', '0002_userprofile_sql_scratchpad'),
    ]

    operations = [
        migrations.RenameField(
            model_name='userprofile',
            old_name='collapse_at',
            new_name='collapse_at_len',
        ),
        migrations.AddField(
            model_name='userprofile',
            name='collapse_at_n_lines',
            field=models.PositiveSmallIntegerField(default=5, verbose_name='Number of lines beyond which result/query field starts collapsed (0 for none)'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/config/apps.py

from django.apps import AppConfig


class ConsentAppConfig(AppConfig):
    name = 'crate_anon.crateweb.consent'


class ResearchAppConfig(AppConfig):
    name = 'crate_anon.crateweb.research'


class UserProfileAppConfig(AppConfig):
    name = 'crate_anon.crateweb.userprofile'


class CoreAppConfig(AppConfig):
    name = 'crate_anon.crateweb.core'



#!/usr/bin/env python
# crate_anon/crateweb/core/management/commands/runcpserver.py

"""
Django management command framework for CherryPy based on:
    https://lincolnloop.com/blog/2008/mar/25/serving-django-cherrypy/
    Idea and code snippets borrowed from
        http://www.xhtml.net/scripts/Django-CherryPy-server-DjangoCerise
    Adapted to run as a management command.
Some bugs fixed by RNC.
Then rewritten by RNC.
Then modified to serve CRATE, with static files, etc.
Then daemonizing code removed:
    https://code.djangoproject.com/ticket/4996

TEST COMMAND:
./manage.py runcpserver --port 8080 --ssl_certificate /etc/ssl/certs/ssl-cert-snakeoil.pem --ssl_private_key /etc/ssl/private/ssl-cert-snakeoil.key
"""  # noqa

from argparse import Namespace
import logging
import sys
# import errno
# import os
# import signal
# import time
# try:
#     import grp
#     import pwd
#     unix = True
# except ImportError:
#     grp = None
#     pwd = None
#     unix = False

import cherrypy
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils import translation

from crate_anon.crateweb.config.wsgi import application as wsgi_application
# COULD ALSO USE:
#   from django.core.handlers.wsgi import WSGIHandler
#   wsgi_application = WSGIHandler()

log = logging.getLogger(__name__)


CRATE_STATIC_URL_PATH = settings.STATIC_URL.rstrip('/')
NEED_UNIX = "Need UNIX for group/user operations"
DEFAULT_ROOT = settings.FORCE_SCRIPT_NAME


class Command(BaseCommand):
    help = ("Run this project in a CherryPy webserver. To do this, "
            "CherryPy is required (pip install cherrypy).")

    def add_arguments(self, parser):
        parser.add_argument(
            '--host', type=str, default="127.0.0.1",
            help="hostname to listen on (default: 127.0.0.1)")
        parser.add_argument(
            '--port', type=int, default=8088,
            help="port to listen on (default: 8088)")
        parser.add_argument(
            "--server_name", type=str, default="localhost",
            help="CherryPy's SERVER_NAME environ entry (default: localhost)")
        # parser.add_argument(
        #     "--daemonize", action="store_true",
        #     help="whether to detach from terminal (default: False)")
        # parser.add_argument(
        #     "--pidfile", type=str,
        #     help="write the spawned process ID to this file")
        # parser.add_argument(
        #     "--workdir", type=str,
        #     help="change to this directory when daemonizing")
        parser.add_argument(
            "--threads", type=int, default=10,
            help="Number of threads for server to use (default: 10)")
        parser.add_argument(
            "--ssl_certificate", type=str,
            help="SSL certificate file "
                 "(e.g. /etc/ssl/certs/ssl-cert-snakeoil.pem)")
        parser.add_argument(
            "--ssl_private_key", type=str,
            help="SSL private key file "
                 "(e.g. /etc/ssl/private/ssl-cert-snakeoil.key)")
        # parser.add_argument(
        #     "--server_user", type=str, default="www-data",
        #     help="user to run daemonized process (default: www-data)")
        # parser.add_argument(
        #     "--server_group", type=str, default="www-data",
        #     help="group to run daemonized process (default: www-data)")
        parser.add_argument(
            "--log_screen", dest="log_screen", action="store_true",
            help="log access requests etc. to terminal (default)")
        parser.add_argument(
            "--no_log_screen", dest="log_screen", action="store_false",
            help="don't log access requests etc. to terminal")
        parser.add_argument(
            "--debug_static", action="store_true",
            help="show debug info for static file requests")
        parser.add_argument(
            "--root_path", type=str, default=DEFAULT_ROOT,
            help="Root path to serve CRATE at. Default: {}".format(
                DEFAULT_ROOT))
        parser.set_defaults(log_screen=True)
        # parser.add_argument(
        #     "--stop", action="store_true",
        #     help="stop server")

    def handle(self, *args, **options):
        opts = Namespace(**options)
        # Activate the current language, because it won't get activated later.
        try:
            translation.activate(settings.LANGUAGE_CODE)
        except AttributeError:
            pass
        # noinspection PyTypeChecker
        runcpserver(opts)


# def change_uid_gid(uid, gid=None):
#     """Try to change UID and GID to the provided values.
#     UID and GID are given as names like 'nobody' not integer.
#
#     Src: http://mail.mems-exchange.org/durusmail/quixote-users/4940/1/
#     """
#     if not unix:
#         raise OSError(NEED_UNIX)
#     if not os.geteuid() == 0:
#         # Do not try to change the gid/uid if not root.
#         return
#     (uid, gid) = get_uid_gid(uid, gid)
#     os.setgid(gid)
#     os.setuid(uid)


# def get_uid_gid(uid, gid=None):
#     """Try to change UID and GID to the provided values.
#     UID and GID are given as names like 'nobody' not integer.
#
#     Src: http://mail.mems-exchange.org/durusmail/quixote-users/4940/1/
#     """
#     if not unix:
#         raise OSError(NEED_UNIX)
#     uid, default_grp = pwd.getpwnam(uid)[2:4]
#     if gid is None:
#         gid = default_grp
#     else:
#         try:
#             gid = grp.getgrnam(gid)[2]
#         except KeyError:
#             gid = default_grp
#     return uid, gid


# def still_alive(pid):
#     """
#     Poll for process with given pid up to 10 times waiting .25 seconds in
#     between each poll.
#     Returns False if the process no longer exists otherwise, True.
#     """
#     for n in range(10):
#         time.sleep(0.25)
#         try:
#             # poll the process state
#             os.kill(pid, 0)
#         except OSError as e:
#             if e[0] == errno.ESRCH:
#                 # process has died
#                 return False
#             else:
#                 raise  # TODO
#     return True


# def stop_server(pidfile):
#     """
#     Stop process whose pid was written to supplied pidfile.
#     First try SIGTERM and if it fails, SIGKILL.
#     If process is still running, an exception is raised.
#     """
#     if os.path.exists(pidfile):
#         pid = int(open(pidfile).read())
#         try:
#             os.kill(pid, signal.SIGTERM)
#         except OSError:  # process does not exist
#             os.remove(pidfile)
#             return
#         if still_alive(pid):
#             # process didn't exit cleanly, make one last effort to kill it
#             os.kill(pid, signal.SIGKILL)
#             if still_alive(pid):
#                 raise OSError("Process {} did not stop.".format(pid))
#         os.remove(pidfile)


class Missing(object):
    """Basic web interface to say "not here"."""
    config = {
        '/': {
            # Anything so as to prevent complaints about an empty config.
            'tools.sessions.on': False,
        }
    }

    @cherrypy.expose
    def index(self):
        return "[CherryPy server says:] Nothing to see here. Wrong URL path."


def start_server(opts):
    """
    Start CherryPy server
    """

    # if opts.daemonize and opts.server_user and opts.server_group:
    #     # ensure the that the daemon runs as specified user
    #     change_uid_gid(opts.server_user, opts.server_group)

    cherrypy.config.update({
        'server.socket_host': opts.host,
        'server.socket_port': opts.port,
        'server.thread_pool': opts.threads,
        'server.server_name': opts.server_name,
        'server.log_screen': opts.log_screen,
    })
    if opts.ssl_certificate and opts.ssl_private_key:
        cherrypy.config.update({
            'server.ssl_module': 'builtin',
            'server.ssl_certificate': opts.ssl_certificate,
            'server.ssl_private_key': opts.ssl_private_key,
        })

    log.info("Starting on host: {}".format(opts.host))
    log.info("Starting on port: {}".format(opts.port))
    log.info("Static files will be served from filesystem path: {}".format(
        settings.STATIC_ROOT))
    log.info("Static files will be served at URL path: {}".format(
        CRATE_STATIC_URL_PATH))
    log.info("CRATE will be at: {}".format(opts.root_path))
    log.info("Thread pool size: {}".format(opts.threads))

    static_config = {
        '/': {
            'tools.staticdir.root': settings.STATIC_ROOT,
            'tools.staticdir.debug': opts.debug_static,
        },
        CRATE_STATIC_URL_PATH: {
            'tools.staticdir.on': True,
            'tools.staticdir.dir': '',
        },
    }
    cherrypy.tree.mount(Missing(), '', config=static_config)
    cherrypy.tree.graft(wsgi_application, opts.root_path)

    try:
        cherrypy.engine.start()
        cherrypy.engine.block()
    except KeyboardInterrupt:
        cherrypy.engine.stop()


def runcpserver(opts):
    # if opts.stop:
    #     if not opts.pidfile:
    #         raise ValueError("Must specify --pidfile to use --stop")
    #     print('stopping server')
    #     stop_server(opts.pidfile)
    #     return True

    # if opts.daemonize:
    #     if not opts.pidfile:
    #         opts.pidfile = '/var/run/cpserver_{}.pid'.format(opts.port)
    #     stop_server(opts.pidfile)
    #
    #     if opts.workdir:
    #         become_daemon(our_home_dir=opts.workdir)
    #     else:
    #         become_daemon()
    #
    #     fp = open(opts.pidfile, 'w')
    #     fp.write("{}\n".format(os.getpid()))
    #     fp.close()

    # Start the webserver
    log.info('starting server with options {}'.format(opts))
    start_server(opts)


if __name__ == '__main__':
    runcpserver(sys.argv[1:])



#!/usr/bin/env python
# crate_anon/anonymise/anon_constants.py

"""
Shared constants for CRATE anonymiser.

Author: Rudolf Cardinal
Created at: 18 Feb 2015
Last update: 22 Nov 2015

Copyright/licensing:

    Copyright (C) 2015-2016 Rudolf Cardinal (rudolf@pobox.com).
    Department of Psychiatry, University of Cambridge.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

from enum import unique

from sqlalchemy import BigInteger, Integer
from cardinal_pythonlib.rnc_lang import StrEnum

from crate_anon.version import VERSION, VERSION_DATE

# =============================================================================
# Logging
# =============================================================================

LOG_FORMAT = '%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:%(message)s'
LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'

LOG_COLORS = {
    'DEBUG': 'cyan',
    'INFO': 'green',
    'WARNING': 'yellow',
    'ERROR': 'red',
    'CRITICAL': 'red,bg_white',
}

# =============================================================================
# Cosmetic
# =============================================================================

BIGSEP = "=" * 20 + " "
SEP = "-" * 20 + " "

# =============================================================================
# Defaults for command-line options
# =============================================================================

DEFAULT_REPORT_EVERY = 100000
DEFAULT_CHUNKSIZE = 1000000

# =============================================================================
# Environment
# =============================================================================

CONFIG_ENV_VAR = 'CRATE_ANON_CONFIG'

# =============================================================================
# Data dictionary
# =============================================================================

DATEFORMAT_ISO8601 = "%Y-%m-%dT%H:%M:%S%z"  # e.g. 2013-07-24T20:04:07+0100
DEFAULT_INDEX_LEN = 20  # for data types where it's mandatory
DEFAULT_MAX_ROWS_BEFORE_COMMIT = 1000
DEFAULT_MAX_BYTES_BEFORE_COMMIT = 80 * 1024 * 1024

LONGTEXT = "LONGTEXT"

MAX_PID_STR = "9" * 10  # e.g. NHS numbers are 10-digit

# Better overall than string.maketrans:
ODD_CHARS_TRANSLATE = [chr(x) for x in range(0, 256)]
for c in '()/ ':
    ODD_CHARS_TRANSLATE[ord(c)] = '_'
for i in range(0, 32):
    ODD_CHARS_TRANSLATE[i] = '_'
for i in range(127, 256):
    ODD_CHARS_TRANSLATE[i] = '_'
ODD_CHARS_TRANSLATE = "".join(ODD_CHARS_TRANSLATE)

PidType = BigInteger
TridType = Integer
MAX_TRID = 2 ** 31 - 1
# https://dev.mysql.com/doc/refman/5.0/en/numeric-type-overview.html
# Maximum INT UNSIGNED is              4294967295 == 2 ** 32 - 1.
# INT range is                        -2147483648 == -(2 **  31) to
#                                     +2147483647 == 2 ** 31 - 1
# ... note that this is inadequate for 10-digit NHS numbers.
# Maximum BIGINT UNSIGNED is 18446744073709551615 == 2 ** 64 - 1.
# BIGINT range is            -9223372036854775808 == -(2 ** 63) to
#                            +9223372036854775807 == 2 ** 64 - 1


@unique
class ALTERMETHOD(StrEnum):
    TRUNCATEDATE = "truncate_date"
    SCRUBIN = "scrub"
    BIN2TEXT = "binary_to_text"
    FILENAME2TEXT = "filename_to_text"
    SKIP_IF_TEXT_EXTRACT_FAILS = "skip_if_extract_fails"
    # HTML_ESCAPE = "html_escape"
    HTML_UNESCAPE = "html_unescape"
    HTML_UNTAG = "html_untag"


@unique
class DECISION(StrEnum):
    OMIT = "OMIT"
    INCLUDE = "include"


@unique
class INDEX(StrEnum):
    NORMAL = "I"
    UNIQUE = "U"
    FULLTEXT = "F"


@unique
class SCRUBMETHOD(StrEnum):
    WORDS = "words"
    PHRASE = "phrase"
    NUMERIC = "number"
    DATE = "date"
    CODE = "code"


@unique
class SCRUBSRC(StrEnum):
    PATIENT = "patient"
    THIRDPARTY = "thirdparty"
    THIRDPARTY_XREF_PID = "thirdparty_xref_pid"


@unique
class SRCFLAG(StrEnum):
    PK = "K"
    ADD_SRC_HASH = "H"
    PRIMARY_PID = "P"
    DEFINES_PRIMARY_PIDS = "*"
    MASTER_PID = "M"
    CONSTANT = "C"
    ADDITION_ONLY = "A"
    OPT_OUT = "!"
    REQUIRED_SCRUBBER = "R"


# =============================================================================
# Databases
# =============================================================================

MYSQL_CHARSET = 'utf8'
MYSQL_TABLE_KWARGS = {
    'mysql_charset': MYSQL_CHARSET,
    'mysql_engine': 'InnoDB',
}

MYSQL_MAX_IDENTIFIER_LENGTH = 64
# http://dev.mysql.com/doc/refman/5.7/en/identifiers.html

# =============================================================================
# Demo config
# =============================================================================

# noinspection PyPep8
DEMO_CONFIG = """# Configuration file for CRATE anonymiser (crate_anonymise).
# Version {VERSION} ({VERSION_DATE}).
#
# Boolean values can be 0/1, Y/N, T/F, True/False.

# =============================================================================
# Main settings
# =============================================================================

[main]

# -----------------------------------------------------------------------------
# Data dictionary
# -----------------------------------------------------------------------------
# Specify a data dictionary in TSV (tab-separated value) format, with a header
# row.
# Columns in the data dictionary (which can be in any order as long as the
# header row matches the data):
#
# src_db
#     Specify the source database.
#     Database names are those used in source_databases list below; they
#     don't have to be SQL database names.
# src_table
#     Table name in source database.
# src_field
#     Field name in source database.
# src_datatype
#     SQL data type in source database, e.g. INT, VARCHAR(50).
#
# src_flags
#     One or more of the following characters:
#
#     {SRCFLAG.PK}
#         PK
#         This field is the primary key (PK) for the table it's in.
#
#     {SRCFLAG.ADD_SRC_HASH}
#         ADD SOURCE HASH
#         Add source hash of the record, for incremental updates?
#         - This flag may only be set for src_pk fields (which cannot then be
#           omitted in the destination, and which require the index={INDEX.UNIQUE}
#           setting, so that a unique index is created for this field).
#         - If set, a field is added to the destination table, with field
#           name as set by the config's source_hash_fieldname variable,
#           containing a hash of the contents of the source record -- all
#           fields that are not omitted, OR contain scrubbing information
#           (scrub_src). The field is of type VARCHAR and its length is
#           determined by the hash_method parameter (see below).
#         - This table is then capable of incremental updates.
#
#     {SRCFLAG.CONSTANT}
#         CONSTANT
#         Record contents are constant (will not change) for a given PK.
#         - An alternative to '{SRCFLAG.ADD_SRC_HASH}'. Can't be used with it.
#         - The flag can be set only on src_pk fields, which can't be omitted
#           in the destination, and which have the same index requirements as
#           the '{SRCFLAG.ADD_SRC_HASH}' flag.
#         - If set, no hash is added to the destination, but the destination
#           contents are assumed to exist and not to have changed.
#         - Be CAUTIOUS with this flag, i.e. certain that the contents will
#           not change.
#         - Intended for very data-intensive fields, such as BLOB fields
#           containing binary documents, where hashing would be quite slow
#           over many gigabytes of data.
#         - Does not imply that the whole table cannot change!
#
#     {SRCFLAG.ADDITION_ONLY}
#         ADDITION ONLY
#         Addition only. It is assumed that records can only be added, not
#         deleted.
#
#     {SRCFLAG.PRIMARY_PID}
#         PRIMARY PID
#         Primary patient ID field. If set,
#         (a) This field will be used to link records for the same patient
#             across all tables. It must therefore be present, and marked in
#             the data dictionary, for ALL tables that contain patient-
#             identifiable information.
#         (b) If the field is not omitted: the field will be hashed as the
#             primary ID (database patient primary key) in the destination,
#             and a transient research ID (TRID) also added.
#
#     {SRCFLAG.DEFINES_PRIMARY_PIDS}
#         DEFINES PRIMARY PIDS.
#         This field *defines* primary PIDs. If set, this row will be used to
#         search for all patient IDs, and will define them for this database.
#         Only those patients will be processed (for all tables containing
#         patient info). Typically, this flag is applied to a SINGLE field in a
#         SINGLE table, usually the principal patient registration/demographics
#         table.
#
#     {SRCFLAG.MASTER_PID}
#         MASTER PID
#         Master ID (e.g. NHS number).
#         The field will be hashed with the master PID hasher.
#
#     {SRCFLAG.OPT_OUT}
#         OPT OUT
#         This field is used to mark that the patient wishes to opt out
#         entirely. It must be in a field that also has a primary patient ID
#         field (because that's the ID that will be omitted). If the opt-out
#         field contains a value that's defined in the optout_col_values
#         setting (see below), that patient will be opted out entirely from
#         the anonymised database.
#
#     {SRCFLAG.REQUIRED_SCRUBBER}
#         REQUIRED SCRUBBER
#         If this field is a scrub_src field (see below), and this flag is set,
#         then at least one non-NULL value for this field must be present for
#         each patient, or no information will be processed for this patient.
#         (Typical use: where you have a master patient index separate from the
#         patient name table, and data might have been brought across
#         partially, so there are some missing names. In this situation, text
#         might go unscrubbed because the names are missing. Setting this flag
#         for the name field will prevent this.)
#
# scrub_src
#     One of:
#           "{SCRUBSRC.PATIENT}",
#           "{SCRUBSRC.THIRDPARTY}",
#           "{SCRUBSRC.THIRDPARTY_XREF_PID}",
#           or blank.
#     Explanations:
#     - "{SCRUBSRC.PATIENT}":
#       Contains patient-identifiable information that must be removed from
#       "scrub_in" fields.
#     - "{SCRUBSRC.THIRDPARTY}":
#       Contains identifiable information about carer/family/other third party,
#       which must be removed from "scrub_in" fields.
#     - "{SCRUBSRC.THIRDPARTY_XREF_PID}":
#       This field is a patient identifier for ANOTHER patient (such as a
#       relative). The scrubber should recursively include THAT patient's
#       identifying information as third-party information for THIS patient.
#
# scrub_method
#     Applicable to scrub_src fields. Manner in which this field should be
#     treated for scrubbing.
#     Options:
#
#     - "{SCRUBMETHOD.WORDS}"
#       Treat as a set of textual words. This is the default for all textual
#       fields (e.g. CHAR, VARCHAR, TEXT). Typically used for names.
#       Also OK for e-mail addresses.
#
#     - "{SCRUBMETHOD.PHRASE}"
#       Treat as a textual phrase (a sequence of words to be replaced only when
#       they occur in sequence). Typically used for address components.
#
#     - "{SCRUBMETHOD.NUMERIC}"
#       Treat as a number. This is the default for all numeric fields (e.g.
#       INTEGER, FLOAT). If you have a phone number in a text field, use this
#       method; it will be scrubbed regardless of spacing/punctuation.
#
#     - "{SCRUBMETHOD.CODE}"
#       Teat as an alphanumeric code. Suited to postcodes. Very like the
#       numeric method, but permits non-digits.
#
#     - "{SCRUBMETHOD.DATE}"
#       Treat as a date. This is the default for all DATE/DATETIME fields.
#
# decision
#     One of:
#     - "{DECISION.OMIT}": omit the field from the output entirely;
#     - "{DECISION.INCLUDE}": include it.
#     This is case sensitive, for safety.
#
# inclusion_values
#     - Either blank, or an expression that evaluates to a Python iterable
#       (e.g. list or tuple) with Python's ast.literal_eval() function (see
#       https://docs.python.org/3.4/library/ast.html).
#     - Examples:
#           [None, 0]
#           [True, 1, 'yes', 'true', 'Yes', 'True']
#     - If this is not blank/None, then it serves as a ROW INCLUSION LIST -
#       the source row will only be processed if the field's value is one of
#       the inclusion values.
#     - It applies to the raw value from the database (before any
#       transformation via alter_method).
#     - This is not applied to scrub_src fields (which contribute to the
#       scrubber regardless.
#     - Note that "[None]" is a list with one member, None, whereas "None"
#       is equivalent to leaving the field blank.
#
# exclusion_values
#     - As for inclusion_values, but the row is excluded if the field's value
#       is in the exclusion_values list.
#
# alter_method
#     Manner in which to alter the data. Blank, or one or more of:
#
#     - "{ALTERMETHOD.SCRUBIN}"
#       Scrub in. Applies to text fields only. The field will have its contents
#       anonymised (using information from other fields). Use this for any
#       text field that end users might store free-text comments in.
#
#     - "{ALTERMETHOD.TRUNCATEDATE}"
#       Truncate this date to the first of the month. Applicable to text or
#       date-as-text fields.
#
#     - "{ALTERMETHOD.BIN2TEXT}=EXTFIELDNAME"
#       Convert a binary field (e.g. VARBINARY, BLOB) to text (e.g. LONGTEXT).
#       The binary data is taken to be the representation of a document.
#       The field EXTFIELDNAME, which must be in the same source table, must
#       contain the file extension (e.g. "pdf", ".pdf") or a filename with that
#       extension (e.g. "/some/path/mything.pdf"), so that the anonymiser knows
#       how to treat the binary data to extract text from it.
#
#     - "{ALTERMETHOD.FILENAME2TEXT}"
#       As for the binary-to-text option, but the field contains a filename
#       (the contents of which is converted to text), rather than containing
#       binary data directly.
#
#     - "{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS}"
#       If one of the text extraction methods is specified, and this flag is
#       also specified, then the data row will be skipped if text extrcation
#       fails (rather than inserted with a NULL value for the text). This is
#       helpful, for example, if your text-processing pipeline breaks; the
#       option prevents rows being created erroneously with NULL text values,
#       so that a subsequent incremental update will fix the problems once
#       you've fixed your text extraction tools.
#
#     - "{ALTERMETHOD.HTML_UNESCAPE}"
#       HTML encoding is removed, e.g. convert "&amp;" to "&" and "&lt;" to "<"
#
#     - "{ALTERMETHOD.HTML_UNTAG}"
#       HTML tags are removed, e.g. from
#           <a href="http://somewhere">see link</a>
#       to
#           see link
#
#     You can specify multiple options separated by commas.
#     Not all are compatible (e.g. scrubbing is for text; date truncation is
#     for dates).
#     If there's more than one, text extraction from BLOBs/files is performed
#     first. After that, they are executed in sequence. (The position of the
#     skip-if-text-extraction-fails flag is immaterial.)
#     A typical combination might be:
#           {ALTERMETHOD.FILENAME2TEXT},{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS},{ALTERMETHOD.SCRUBIN}
#     or:
#           {ALTERMETHOD.HTML_UNTAG},{ALTERMETHOD.HTML_UNESCAPE},{ALTERMETHOD.SCRUBIN}
#
# dest_table
#     Table name in destination database.
# dest_field
#     Field name in destination database.
# dest_datatype
#     SQL data type in destination database.
#     If omitted, the source SQL data type is translated appropriately.
# index
#     One of:
#     - blank: no index.
#     - "{INDEX.NORMAL}"
#       ... create a normal index on the destination field.
#     - "{INDEX.UNIQUE}"
#       ... create a unique index on the destination field.
#     - "{INDEX.FULLTEXT}"
#       ... create a FULLTEXT index, for rapid searching within long text
#       fields. Only applicable to one field per table.
# indexlen
#     Integer. Can be blank. If not, sets the prefix length of the index.
#     Mandatory in MySQL if you apply a normal (+/- unique) index to a TEXT
#     or BLOB field. Not required for FULLTEXT indexes.
# comment
#     Field comment, stored in destination database.

data_dictionary_filename = testdd.tsv

# -----------------------------------------------------------------------------
# Encryption phrases/passwords
# -----------------------------------------------------------------------------

    # PID-to-RID hashing method. Options are:
    # - HMAC_MD5 - produces a 32-character digest
    # - HMAC_SHA256 - produces a 64-character digest
    # - HMAC_SHA512 - produces a 128-character digest
hash_method = HMAC_MD5

per_table_patient_id_encryption_phrase = SOME_PASSPHRASE_REPLACE_ME

master_patient_id_encryption_phrase = SOME_OTHER_PASSPHRASE_REPLACE_ME

change_detection_encryption_phrase = YETANOTHER

# -----------------------------------------------------------------------------
# Text extraction
# -----------------------------------------------------------------------------

    # Use the plainest possible layout for text extraction?
    # False = better for human layout. Table example from DOCX:
    #     +---------+---------+
    #     | AAA AAA | BBB BBB |
    #     | AAA AAA | BBB BBB |
    #     +---------+---------+
    # True = good for natural language processing. Table example from DOCX:
    #     ---------------------
    #       AAA AAA
    #       AAA AAA
    #     ---------------------
    #                 BBB BBB
    #                 BBB BBB
    #     ---------------------
    # ... note the absence of vertical interruptions, and that text from one
    # cell remains contiguous.
extract_text_plain = False

    # Default width to word-wrap extracted text to
extract_text_width = 80

# -----------------------------------------------------------------------------
# Anonymisation
# -----------------------------------------------------------------------------

    # Patient information will be replaced with this. For example, XXXXXX or
    # [___] or [__PPP__] or [__ZZZ__]; the bracketed forms can be a bit easier
    # to spot, and work better if they directly abut other text.
replace_patient_info_with = [__PPP__]

    # Third-party information will be replaced by this.
    # For example, YYYYYY or [...] or [__TTT__] or [__QQQ__].
replace_third_party_info_with = [__TTT__]

    # For fields marked as scrub_src = {SCRUBSRC.THIRDPARTY_XREF_PID},
    # how deep should we recurse? The default is 1. Beware making this too
    # large; the recursion trawls a lot of information (and also uses an
    # extra simultaneous database cursor for each recursion).
thirdparty_xref_max_depth = 1

    # Things to be removed irrespective of patient-specific information will be
    # replaced by this (for example, if you opt to remove all things looking
    # like telephone numbers). For example, ZZZZZZ or [~~~].
replace_nonspecific_info_with = [~~~]

    # Strings to append to every "scrub from" string.
    # For example, include "s" if you want to scrub "Roberts" whenever you
    # scrub "Robert".
    # Applies to {SCRUBMETHOD.WORDS}, but not to {SCRUBMETHOD.PHRASE}.
    # Multiline field: https://docs.python.org/2/library/configparser.html
scrub_string_suffixes =
    s

    # Specify maximum number of errors (insertions, deletions, substitutions)
    # in string regex matching. Beware using a high number! Suggest 1-2.
string_max_regex_errors = 1

    # Is there a minimum length to apply string_max_regex_errors? For example,
    # if you allow one typo and someone is called Ian, all instances of 'in' or
    # 'an' will be wiped. Note that this apply to scrub-source data.
min_string_length_for_errors = 4

    # Is there a minimum length of string to scrub WITH? For example, if you
    # specify 2, you allow two-letter names such as Al to be scrubbed, but you
    # allow initials through, and therefore prevent e.g. 'A' from being
    # scrubbed from the destination. Note that this applies to scrub-source
    # data.
min_string_length_to_scrub_with = 2

    # WHITELIST.
    # Are there any words not to scrub? For example, "the", "road", "street"
    # often appear in addresses, but you might not want them removed. Be
    # careful in case these could be names (e.g. "Lane").
    # Specify these as a list of FILENAMES, where the files contain words; e.g.
    #
    # whitelist_filenames = /some/path/short_english_words.txt
    #
    # Here's a suggestion for some of the sorts of words you might include:
    #     am
    #     an
    #     as
    #     at
    #     bd
    #     by
    #     he
    #     if
    #     is
    #     it
    #     me
    #     mg
    #     od
    #     of
    #     on
    #     or
    #     re
    #     so
    #     to
    #     us
    #     we
    #     her
    #     him
    #     tds
    #     she
    #     the
    #     you
    #     road
    #     street
whitelist_filenames =

    # BLACKLIST
    # Are there any words you always want to remove?
    # Specify these as a list of filenames, e.g
    #
    # blacklist_filenames = /some/path/boy_names.txt
    #     /some/path/girl_names.txt
    #     /some/path/common_surnames.txt
blacklist_filenames =

    # Nonspecific scrubbing of numbers of a certain length?
    # For example, scrubbing all 11-digit numbers will remove modern UK
    # telephone numbers in conventional format. To do this, specify
    # scrub_all_numbers_of_n_digits = 11. You could scrub both 10- and 11-digit
    # numbers by specifying both numbers (in multiline format, as above);
    # 10-digit numbers would include all NHS numbers. Avoid using this for
    # short numbers; you may lose valuable numeric data!
scrub_all_numbers_of_n_digits =

    # Nonspecific scrubbing of UK postcodes?
    # See https://www.mrs.org.uk/pdf/postcodeformat.pdf ; these can look like
    # FORMAT    EXAMPLE
    # AN NAA    M1 1AA
    # ANN NAA   M60 1NW
    # AAN NAA   CR2 6XH
    # AANN NAA  DN55 1PT
    # ANA NAA   W1A 1HQ
    # AANA NAA  EC1A 1BB
scrub_all_uk_postcodes = False

    # Anonymise at word boundaries? True is more conservative; False is more
    # liberal and will deal with accidental word concatenation. With ID
    # numbers, beware if you use a prefix, e.g. if people write 'M123456' or
    # 'R123456'; in that case you will need
    #       anonymise_numbers_at_word_boundaries_only = False.
anonymise_codes_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.CODE}
anonymise_dates_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.DATE}
anonymise_numbers_at_word_boundaries_only = False
    # ... applies to {SCRUBMETHOD.NUMERIC}
anonymise_numbers_at_numeric_boundaries_only = True
    # ... applies to {SCRUBMETHOD.NUMERIC}
    # ... if True, will not scrub "234" from "123456"
    # ... setting this to False is extremely conservative
anonymise_strings_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.WORDS} and {SCRUBMETHOD.PHRASE}

# -----------------------------------------------------------------------------
# Output fields and formatting
# -----------------------------------------------------------------------------

    # Name used for the primary patient ID in the mapping table.
mapping_patient_id_fieldname = patient_id

    # Research ID field name. This will be a VARCHAR of length determined by
    # hash_method. Used to replace per_table_patient_id_field.
research_id_fieldname = brcid

    # Transient integer research ID (TRID) fieldname.
    # An unsigned integer field with this name will be added to every table
    # containing a primary patient ID (in the source) or research ID (in the
    # destination).
trid_fieldname = trid

    # Name used for the master patient ID in the mapping table.
mapping_master_id_fieldname = nhsnum

    # Similarly, used to replace ddgen_master_pid_fieldname:
master_research_id_fieldname = nhshash

    # Change-detection hash fieldname. This will be a VARCHAR of length
    # determined by hash_method.
source_hash_fieldname = _src_hash

    # Date-to-text conversion formats
date_to_text_format = %Y-%m-%d
    # ... ISO-8601, e.g. 2013-07-24
datetime_to_text_format = %Y-%m-%dT%H:%M:%S
    # ... ISO-8601, e.g. 2013-07-24T20:04:07

    # Append source table/field to the comment? Boolean.
append_source_info_to_comment = True

# -----------------------------------------------------------------------------
# Destination database configuration
# See the [destination_database] section for connection details.
# -----------------------------------------------------------------------------

    # Specify the maximum number of rows to be processed before a COMMIT is
    # issued on the database transaction. This prevents the transaction growing
    # too large.
    # Default is {DEFAULT_MAX_ROWS_BEFORE_COMMIT}.
max_rows_before_commit = {DEFAULT_MAX_ROWS_BEFORE_COMMIT}

    # Specify the maximum number of source-record bytes (approximately!) that
    # are processed before a COMMIT is issued on the database transaction. This
    # prevents the transaction growing too large. The COMMIT will be issued
    # *after* this limit has been met/exceeded, so it may be exceeded if the
    # transaction just before the limit takes the cumulative total over the
    # limit.
    # Default is {DEFAULT_MAX_BYTES_BEFORE_COMMIT}.
max_bytes_before_commit = {DEFAULT_MAX_BYTES_BEFORE_COMMIT}

    # We need a temporary table name for incremental updates. This can't be the
    # name of a real destination table. It lives in the destination database.
temporary_tablename = _temp_table

# -----------------------------------------------------------------------------
# Choose databases (defined in their own sections).
# -----------------------------------------------------------------------------

    # Source database list. Can be lots.
source_databases =
    mysourcedb1
    mysourcedb2

    # Destination database. Just one.
destination_database = my_destination_database

    # Admin database. Just one.
admin_database = my_admin_database

# -----------------------------------------------------------------------------
# PROCESSING OPTIONS, TO LIMIT DATA QUANTITY FOR TESTING
# -----------------------------------------------------------------------------

    # Limit the number of patients to be processed? Specify 0 (the default) for
    # no limit.
debug_max_n_patients =

    # Specify a list of integer patient IDs, for debugging? If specified, this
    # list will be used directly (overriding the patient ID source specified in
    # the data dictionary, and overriding debug_max_n_patients).
debug_pid_list =

# =============================================================================
# Opting out entirely
# =============================================================================
# Patients who elect to opt out entirely have their PIDs stored in the OptOut
# table of the admin database. ENTRIES ARE NEVER REMOVED FROM THIS LIST BY
# CRATE. It can be populated in three ways:
#   1. Manually, by adding a PID to the column opt_out.pid).
#   2. By maintaining a text file list of integer PIDs. Any PIDs in this file
#      are added to the opt-out list.
#   3. By flagging a source database field as indicating an opt-out, using the
#      src_flags = "{SRCFLAG.OPT_OUT}" marker.

    # If you set this, each line of the file(s) is scanned for an integer,
    # taken to the PID of a patient who wishes to opt out.
optout_pid_filenames =

    # If you set this, each line of the file(s) is scanned for an integer,
    # taken to the MPID of a patient who wishes to opt out.
optout_mpid_filenames =

    # If you mark a field in the data dictionary as an opt-out field (see
    # above), that says "the field tells you whether the patient opts out or
    # not". But is it "opt out" or "not"? If the actual value matches one
    # below, then it's "opt out". Specify a LIST OF PYTHON VALUES; for example:
    #       optout_col_values = [True, 'Yes', 'Y']
optout_col_values =

# =============================================================================
# Destination database details. User should have WRITE access.
# =============================================================================
# Use SQLAlchemy URLs: see
#       http://docs.sqlalchemy.org/en/latest/core/engines.html
# You may need to install additional drivers, e.g.
#       pip install SOME_DRIVER
# ... see the documentation.

[my_destination_database]

url = mysql+mysqldb://username:password@127.0.0.1:3306/output_databasename?charset=utf8

# =============================================================================
# Administrative database. User should have WRITE access.
# =============================================================================

[my_admin_database]

url = mysql+mysqldb://username:password@127.0.0.1:3306/admin_databasename?charset=utf8

# =============================================================================
# SOURCE DATABASE DETAILS BELOW HERE.
# User should have READ access only for safety.
# =============================================================================

# -----------------------------------------------------------------------------
# Source database example 1
# -----------------------------------------------------------------------------

[mysourcedb1]

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # CONNECTION DETAILS
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

url = mysql+mysqldb://username:password@127.0.0.1:3306/source_databasename?charset=utf8

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # INPUT FIELDS, FOR THE AUTOGENERATION OF DATA DICTIONARIES
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # - For field specifications, fields can either be specified as "column"
    #   (to match any table) or "table.column", to match a specific table.
    #   They are case-insensitive.
    #   Wildcards (*, ?) may also be used (as per Python's fnmatch).

    # By default, most fields (except PKs and patient ID codes) are marked
    # as "OMIT", pending human review. If you want to live dangerously, set
    # this to False, and they will be marked as "include" from the outset.
ddgen_omit_by_default = True

    # You can specify additional fields to omit...
ddgen_omit_fields =

    # ... or include. "Omit" overrides "include".
    # If a field contains scrubbing source information, it will also be omitted
    # pending human review, regardless of other settings.
ddgen_include_fields =

    # Allow the absence of patient info? Used to copy databases; WILL NOT
    # ANONYMISE. Boolean; default is False.
ddgen_allow_no_patient_info = False

    # Specify the (typically integer) patient identifier present in EVERY
    # table. It will be replaced by the research ID in the destination
    # database.
ddgen_per_table_pid_field = patient_id

    # Add every instance of a per-table PID field to the patient scrubber?
    # This is a very conservative setting, and should be unnecessary as the
    # single master "PID-defining" column (see ddgen_pid_defining_fieldnames)
    # should be enough.
    # (Note that per-table PIDs are always replaced by RIDs - this setting
    # governs whether the scrubber used to scrub free-text fields also
    # works through every single per-table PID).
ddgen_add_per_table_pids_to_scrubber = False

    # Master patient ID fieldname. Used for e.g. NHS numbers.
ddgen_master_pid_fieldname = nhsnum

    # Blacklist any tables when creating new data dictionaries?
    # This is case-insensitive, and you can use */? wildcards (as per Python's
    # fnmatch module).
ddgen_table_blacklist =

    # Whitelist any tables? (Whitelists override blacklists.)
ddgen_table_whitelist =

    # Blacklist any fields (regardless of their table) when creating new data
    # dictionaries? Wildcards of */? operate as above.
ddgen_field_blacklist =

    # Whitelist any fields? (Whitelists override blacklists.)
ddgen_field_whitelist =

    # Fieldnames assumed to be their table's PK:
ddgen_pk_fields =

    # Assume that content stays constant?
    # (Applies {SRCFLAG.CONSTANT} to PK fields; q.v.)
    # This is the default; then ddgen_constant_content_tables and
    # ddgen_nonconstant_content_tables can override (of which,
    # ddgen_nonconstant_content_tables takes priority if a table matches both).
ddgen_constant_content = False

    # Table-specific overrides for ddgen_constant_content, as above.
ddgen_constant_content_tables =
ddgen_nonconstant_content_tables =

    # Assume that records can only be added, not deleted?
ddgen_addition_only = False

    # Table-specific overrides for ddgen_addition_only, similarly.
ddgen_addition_only_tables =
ddgen_deletion_possible_tables =

    # Predefine field(s) that define the existence of patient IDs? UNUSUAL.
ddgen_pid_defining_fieldnames =

    # Default fields to scrub from
ddgen_scrubsrc_patient_fields =
ddgen_scrubsrc_thirdparty_fields =
ddgen_scrubsrc_thirdparty_xref_pid_fields =

    # Are any scrub_src fields required (mandatory), i.e. must have non-NULL
    # data in at least one row?
ddgen_required_scrubsrc_fields =

    # Override default scrubbing methods
ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields =
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =

    # Known safe fields, exempt from scrubbing
ddgen_safe_fields_exempt_from_scrubbing =

    # Define minimum text field length for scrubbing (shorter is assumed safe)
ddgen_min_length_for_scrubbing = 4

    # Other default manipulations
ddgen_truncate_date_fields =

    # Fields containing filenames, which files should be converted to text
ddgen_filename_to_text_fields =

    # Fields containing raw binary data from files (binary large objects;
    # BLOBs), whose contents should be converted to text -- paired with fields
    # in the same table containing their file extension (e.g. "pdf", ".PDF") or
    # a filename having that extension.
    # Specify it as a list of comma-joined pairs, e.g.
    #     ddgen_binary_to_text_field_pairs = binary1field, ext1field
    #         binary2field, ext2field
    #         ...
    # The first (binaryfield) can be specified as column or table.column,
    # but the second must be column only.
ddgen_binary_to_text_field_pairs =

    # Specify any text-extraction rows for which you also want to set the flag
    # "{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS}":
ddgen_skip_row_if_extract_text_fails_fields =

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # DESTINATION INDEXING
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Fields to apply an index to
ddgen_index_fields =

    # Allow full-text index creation? Default true. Disable for databases that
    # don't support them?
ddgen_allow_fulltext_indexing = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # DATA DICTIONARY MANIPULATION TO DESTINATION TABLE/FIELD NAMES
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Force all destination tables/fields to lower case?
    # Boolean; default is True.
ddgen_force_lower_case = True

    # Convert spaces in table/fieldnames (yuk!) to underscores? Default: true.
ddgen_convert_odd_chars_to_underscore = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # PROCESSING OPTIONS, TO LIMIT DATA QUANTITY FOR TESTING
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Specify 0 (the default) for no limit, or a number of rows (e.g. 1000) to
    # apply to any tables listed in debug_limited_tables. For those tables,
    # only this many rows will be taken from the source database. Use this, for
    # example, to reduce the number of large documents fetched.
    # If you run a multiprocess/multithreaded anonymisation, this limit applies
    # per *process* (or task), not overall.
    # Note that these limits DO NOT APPLY to the fetching of patient-
    # identifiable information for anonymisation -- when a patient is
    # processed, all identifiable information for that patient is trawled.
debug_row_limit =

    # List of tables to which to apply debug_row_limit (see above).
debug_limited_tables =

# -----------------------------------------------------------------------------
# Source database example 2
# -----------------------------------------------------------------------------

[mysourcedb2]

url = mysql+mysqldb://username:password@127.0.0.1:3306/source2_databasename?charset=utf8

ddgen_force_lower_case = True
ddgen_per_table_pid_field = patient_id
ddgen_master_pid_fieldname = nhsnum
ddgen_table_blacklist =
ddgen_field_blacklist =
ddgen_pk_fields =
ddgen_constant_content = False
ddgen_scrubsrc_patient_fields =
ddgen_scrubsrc_thirdparty_fields =
ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields =
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =
ddgen_safe_fields_exempt_from_scrubbing =
ddgen_min_length_for_scrubbing = 4
ddgen_truncate_date_fields =
ddgen_filename_to_text_fields =
ddgen_binary_to_text_field_pairs =

# -----------------------------------------------------------------------------
# Source database example 3
# -----------------------------------------------------------------------------

[camcops]
# Example for the CamCOPS anonymisation staging database

url = mysql+mysqldb://username:password@127.0.0.1:3306/camcops_databasename?charset=utf8

# FOR EXAMPLE:
ddgen_force_lower_case = True
ddgen_per_table_pid_field = _patient_idnum1
ddgen_pid_defining_fieldnames = _patient_idnum1
ddgen_master_pid_fieldname = _patient_idnum2

ddgen_table_blacklist =

ddgen_field_blacklist = _patient_iddesc1
    _patient_idshortdesc1
    _patient_iddesc2
    _patient_idshortdesc2
    _patient_iddesc3
    _patient_idshortdesc3
    _patient_iddesc4
    _patient_idshortdesc4
    _patient_iddesc5
    _patient_idshortdesc5
    _patient_iddesc6
    _patient_idshortdesc6
    _patient_iddesc7
    _patient_idshortdesc7
    _patient_iddesc8
    _patient_idshortdesc8
    id
    patient_id
    _device
    _era
    _current
    _when_removed_exact
    _when_removed_batch_utc
    _removing_user
    _preserving_user
    _forcibly_preserved
    _predecessor_pk
    _successor_pk
    _manually_erased
    _manually_erased_at
    _manually_erasing_user
    _addition_pending
    _removal_pending
    _move_off_tablet

ddgen_pk_fields = _pk
ddgen_constant_content = False

ddgen_scrubsrc_patient_fields = _patient_forename
    _patient_surname
    _patient_dob
    _patient_idnum1
    _patient_idnum2
    _patient_idnum3
    _patient_idnum4
    _patient_idnum5
    _patient_idnum6
    _patient_idnum7
    _patient_idnum8

ddgen_scrubsrc_thirdparty_fields =

ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields = _patient_dob
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =

ddgen_safe_fields_exempt_from_scrubbing = _device
    _era
    _when_added_exact
    _adding_user
    _when_removed_exact
    _removing_user
    _preserving_user
    _manually_erased_at
    _manually_erasing_user
    when_last_modified
    when_created
    when_firstexit
    clinician_specialty
    clinician_name
    clinician_post
    clinician_professional_registration
    clinician_contact_details
# ... now some task-specific ones
    bdi_scale
    pause_start_time
    pause_end_time
    trial_start_time
    cue_start_time
    target_start_time
    detection_start_time
    iti_start_time
    iti_end_time
    trial_end_time
    response_time
    target_time
    choice_time
    discharge_date
    discharge_reason_code
    diagnosis_psych_1_icd10code
    diagnosis_psych_1_description
    diagnosis_psych_2_icd10code
    diagnosis_psych_2_description
    diagnosis_psych_3_icd10code
    diagnosis_psych_3_description
    diagnosis_psych_4_icd10code
    diagnosis_psych_4_description
    diagnosis_medical_1
    diagnosis_medical_2
    diagnosis_medical_3
    diagnosis_medical_4
    category_start_time
    category_response_time
    category_chosen
    gamble_fixed_option
    gamble_lottery_option_p
    gamble_lottery_option_q
    gamble_start_time
    gamble_response_time
    likelihood

ddgen_min_length_for_scrubbing = 4

ddgen_truncate_date_fields = _patient_dob
ddgen_filename_to_text_fields =
ddgen_binary_to_text_field_pairs =

""".format(  # noqa
    SCRUBSRC=SCRUBSRC,
    INDEX=INDEX,
    SCRUBMETHOD=SCRUBMETHOD,
    ALTERMETHOD=ALTERMETHOD,
    SRCFLAG=SRCFLAG,
    LONGTEXT=LONGTEXT,
    DEFAULT_MAX_ROWS_BEFORE_COMMIT=DEFAULT_MAX_ROWS_BEFORE_COMMIT,
    DEFAULT_MAX_BYTES_BEFORE_COMMIT=DEFAULT_MAX_BYTES_BEFORE_COMMIT,
    DECISION=DECISION,
    VERSION=VERSION,
    VERSION_DATE=VERSION_DATE,
)

# For the style:
#       [source_databases]
#       source1 = blah
#       source2 = thing
# ... you can't have multiple keys with the same name.
# http://stackoverflow.com/questions/287757



#!/usr/bin/env python
# crate_anon/version.py

"""
Author: Rudolf Cardinal
Created at: 22 Nov 2015
Last update: 9 Mar 2016

Copyright/licensing:

    Copyright (C) 2015-2016 Rudolf Cardinal (rudolf@pobox.com).
    Department of Psychiatry, University of Cambridge.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

VERSION = "0.18.8"
VERSION_DATE = "2016-11-15"



#!/usr/bin/env python
# crate_anon/crateweb/config/settings.py

"""
Django settings for crateweb project.

Generated by 'django-admin startproject' using Django 1.8.4.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

# import datetime
import importlib.machinery
import logging
import os
from crate_anon.crateweb.config.constants import CRATEWEB_CONFIG_ENV_VAR

# http://stackoverflow.com/questions/2636536/how-to-make-django-work-with-unsupported-mysql-drivers-such-as-gevent-mysql-or-c  # noqa
try:
    import pymysql
    pymysql.install_as_MySQLdb()
except ImportError:
    pymysql = None

log = logging.getLogger(__name__)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
# This is the path of the file FROM WHICH THE MODULE WAS LOADED, NOT THE
# DIRECTORY THIS FILE IS IN. So for command-line execute it's e.g.
# /somewhere/crate/crateweb (the directory in which manage.py runs).
# See https://docs.python.org/2/reference/datamodel.html
# Verify this with:

# log.warning("BASE_DIR: {}".format(BASE_DIR))


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.humanize',  # for nice comma formatting of numbers
    'debug_toolbar',  # for debugging
    'django_extensions',  # for graph_models etc.
    'sslserver',  # for SSL testing
    # 'kombu.transport.django',  # for Celery with Django database as broker
    'template_profiler_panel',
    # 'silk',

    'crate_anon.crateweb.config.apps.UserProfileAppConfig',  # for user-specific settings  # noqa
    'crate_anon.crateweb.config.apps.ResearchAppConfig',  # the research database query app  # noqa
    'crate_anon.crateweb.config.apps.ConsentAppConfig',  # the consent-to-contact app  # noqa
    'crate_anon.crateweb.config.apps.CoreAppConfig',  # for e.g. the runcpserver command  # noqa
)

MIDDLEWARE_CLASSES = (
    # 'silk.middleware.SilkyMiddleware',

    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    # Additional:
    'crate_anon.crateweb.extra.middleware.UserBasedExceptionMiddleware',  # provide debugging details to superusers  # noqa
    'crate_anon.crateweb.extra.middleware.LoginRequiredMiddleware',  # prohibit all pages except login pages if not logged in  # noqa
    'crate_anon.crateweb.extra.middleware.DisableClientSideCachingMiddleware',  # no client-side caching  # noqa
    'crate_anon.crateweb.core.middleware.RestrictAdminMiddleware',  # non-developers can't access the devadmin site  # noqa
    # 'crate_anon.crateweb.extra.request_cache.RequestCacheMiddleware',  # per-request cache, UNTESTED  # noqa

    # Last, when using the profiling panel? But actually breaks it...
    # 'debug_toolbar.middleware.DebugToolbarMiddleware',  # should be added automatically, but there's a problem (2016-04-14)  # noqa
)

# Celery things
# BROKER_URL = 'django://'  # for Celery with Django database as broker
BROKER_URL = 'amqp://'  # for Celery with RabbitMQ as broker
CELERY_ACCEPT_CONTENT = ['json']  # avoids potential pickle security problem
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_SERIALIZER = 'json'
# Results are OPTIONAL. The CRATE web service doesn't use them.
# But may be helpful for Celery testing.
# See http://docs.celeryproject.org/en/latest/configuration.html#std:setting-CELERY_RESULT_BACKEND  # noqa
CELERY_RESULT_BACKEND = "rpc://"  # uses AMQP
CELERY_RESULT_PERSISTENT = False

LOGIN_URL = '/login/'  # for LoginRequiredMiddleware
LOGIN_VIEW_NAME = 'login'  # for LoginRequiredMiddleware
LOGIN_EXEMPT_URLS = []  # for LoginRequiredMiddleware

ROOT_URLCONF = 'crate_anon.crateweb.config.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        # 'APP_DIRS': True,  # can't use OPTIONS/loaders with this
        'OPTIONS': {
            'context_processors': [
                # 'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'crate_anon.crateweb.research.context_processors.common_context',  # noqa
            ],
            'loaders': [
                # https://docs.djangoproject.com/en/1.9/ref/templates/api/
                ('django.template.loaders.cached.Loader', [
                    'django.template.loaders.filesystem.Loader',
                    'django.template.loaders.app_directories.Loader',
                ]),
            ],
        },
    },
]

WSGI_APPLICATION = 'crate_anon.crateweb.config.wsgi.application'

DEBUG_TOOLBAR_PANELS = [
    # Standard:
    'debug_toolbar.panels.versions.VersionsPanel',
    'debug_toolbar.panels.timer.TimerPanel',
    'debug_toolbar.panels.settings.SettingsPanel',
    'debug_toolbar.panels.headers.HeadersPanel',
    'debug_toolbar.panels.request.RequestPanel',
    'debug_toolbar.panels.sql.SQLPanel',
    'debug_toolbar.panels.staticfiles.StaticFilesPanel',
    'debug_toolbar.panels.templates.TemplatesPanel',
    'debug_toolbar.panels.cache.CachePanel',
    'debug_toolbar.panels.signals.SignalsPanel',
    'debug_toolbar.panels.logging.LoggingPanel',
    'debug_toolbar.panels.redirects.RedirectsPanel',
    # Built in but not enabled as standard:
    # 'debug_toolbar.panels.profiling.ProfilingPanel',  # EXTREME DANGER! Breaks middleware inc. LoginRequiredMiddleware!  # noqa
    # Extra:
    'template_profiler_panel.panels.template.TemplateProfilerPanel',
]

# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

# USE_L10N = True
USE_L10N = False

# https://docs.djangoproject.com/en/1.8/topics/i18n/timezones/
USE_TZ = True

DATE_FORMAT = "j M Y"  # d for leading zero, j for none
TIME_FORMAT = "H:i"
DATETIME_FORMAT = "d M Y, H:i:s"
# O for timezone e.g. "+0100"

SHORT_DATE_FORMAT = "d/m/Y"
SHORT_TIME_FORMAT = "H:i"
SHORT_DATETIME_FORMAT = "d/m/Y, H:i:s"

# =============================================================================
# Static files (CSS, JavaScript, Images)
# =============================================================================
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = '/crate_static/'
# This is a bit hard-coded, but at least it prevents conflicts with other
# programs. No way to make Django look up (reverse) static URLs dynamically
# using get_script_prefix()?
# (I know that sounds a bit crazy, but the idea would be to point Apache to
# serve those static files via a Django-site-specific location.)
#
# It seems not. So specifying STATIC_URL like this means that Django will serve
# them correctly from its development server, and Apache should be pointed to
# serve from this URL (and from wherever seems best on the filesystem) during
# production.

LOCAL_STATIC_DIR = os.path.join(BASE_DIR, 'static')
STATICFILES_DIRS = (
    LOCAL_STATIC_DIR,
)
STATIC_ROOT = os.path.join(BASE_DIR, 'static_collected')
# ... for collectstatic

# NOTE that deriving from django.contrib.staticfiles.storage.StaticFilesStorage
# and referring to it with STATICFILES_STORAGE only influences the
# "collectstatic" command, not the creation of URLs to static files, so isn't
# relevant here.
# https://docs.djangoproject.com/en/dev/ref/contrib/staticfiles/

# =============================================================================
# Some managed database access goes to the secret mapping database.
# =============================================================================

DATABASE_ROUTERS = ['crate_anon.crateweb.research.models.PidLookupRouter']

# =============================================================================
# Security; https://docs.djangoproject.com/en/1.8/topics/security/
# =============================================================================

CRATE_HTTPS = True  # may be overridden in local settings

SESSION_EXPIRE_AT_BROWSER_CLOSE = True

# =============================================================================
# Logging; https://docs.djangoproject.com/en/1.8/topics/logging/
# =============================================================================

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(name)s:%(levelname)s:%(asctime)s.%(msecs)03d:%(module)s:%(message)s',  # noqa
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'simple': {
            'format': '%(name)s:%(levelname)s:%(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'color': {
            '()': 'colorlog.ColoredFormatter',
            'format': '%(log_color)s%(asctime)s.%(msecs)03d:%(name)s:%(levelname)s:%(message)s',  # noqa
            'datefmt': '%Y-%m-%d %H:%M:%S',
            'log_colors': {
                'DEBUG': 'bold_black',
                'INFO': 'white',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'bold_red',
            },
        }
    },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },
    'handlers': {
        'console': {
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            # 'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
            'formatter': 'color',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
            'formatter': 'verbose',
        },
    },
    'loggers': {
        '': {  # root logger; necessary if everything propagates
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },

        'django': {
            'handlers': ['console', 'mail_admins'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'INFO'),
            'propagate': True,
        },

        # My Django apps:
        'research': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        'consent': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        'core': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        'extra': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },

        # For CherryPy:
        '__main__': {  # for our CherryPy launch script, if used
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        # Not sure the following are working, despite the docs!
        # http://docs.cherrypy.org/en/latest/basics.html#play-along-with-your-other-loggers  # noqa
        'cherrypy_console': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        'cherrypy_access': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
        'cherrypy_error': {
            'handlers': ['console'],
            'level': os.getenv('DJANGO_LOG_LEVEL', 'DEBUG'),
            'propagate': True,
        },
    },
}

# =============================================================================
# PDF generation
# =============================================================================
# https://pypi.python.org/pypi/pdfkit

WKHTMLTOPDF_OPTIONS = {
    # http://wkhtmltopdf.org/usage/wkhtmltopdf.txt
    'page-size': 'A4',
    'dpi': '300',
    'orientation': 'portrait',
    'margin-top': '20mm',
    'margin-right': '20mm',
    'margin-bottom': '20mm',
    'margin-left': '20mm',
    'encoding': 'UTF-8',
    # 'disable-smart-shrinking': None,
    # 'print-media-type': None,
}
PATIENT_FONTSIZE = "11.5pt"  # "12.4pt"
"""
    NB strange wkhtmltopdf bug: word wrap goes awry with font sizes of
      11.6pt to 12.3pt inclusive
    but is fine with
      10, 11, 11.5, 12.4, 12.5, 13, 18...
    Affects addresses and word wrap within tables.
    Using wkhtmltopdf 0.12.2.1 (with patched qt)
    See https://github.com/wkhtmltopdf/wkhtmltopdf/issues/2505
"""
RESEARCHER_FONTSIZE = "10pt"

# =============================================================================
# Import from a site-specific file
# =============================================================================
# First attempt: import file with a fixed name from the PYTHONPATH.
#       from crate_local_settings import *  # noqa
# Better: import a file named in an environment variable.

if ('CRATE_RUN_WITHOUT_LOCAL_SETTINGS' in os.environ and
        os.environ['CRATE_RUN_WITHOUT_LOCAL_SETTINGS'].lower() in
        ['true', '1', 't', 'y', 'yes']):
    log.info("Running without local settings")
    # We will only get here for the collectstatic command in the Debian
    # postinst file, so we just need the minimum specified.
    CLINICAL_LOOKUP_DB = 'dummy_clinical'
    EMAIL_SENDER = 'dummy'
    FORCE_SCRIPT_NAME = ''
    MAX_UPLOAD_SIZE_BYTES = 1000
    PRIVATE_FILE_STORAGE_ROOT = '/tmp'
    RESEARCH_DB_TITLE = "dummy database title"
    SECRET_KEY = 'dummy'
    SECRET_MAP = {
        'TABLENAME': 'dummy',
        'PID_FIELD': 'dummy',
        'RID_FIELD': 'dummy',
        'MASTER_PID_FIELD': 'dummy',
        'MASTER_RID_FIELD': 'dummy',
        'TRID_FIELD': 'dummy',
        'MAX_RID_LENGTH': 255,
    }
else:
    if CRATEWEB_CONFIG_ENV_VAR not in os.environ:
        raise ValueError("""
    You must set the {e} environment variable first.
    Aim it at your settings file, like this:

    export {e}=/etc/crate/my_secret_crate_settings.py
        """.format(e=CRATEWEB_CONFIG_ENV_VAR))
    log.info("Loading local settings")
    _loader = importlib.machinery.SourceFileLoader(
        'local_settings',
        os.environ[CRATEWEB_CONFIG_ENV_VAR])
    _local_module = _loader.load_module()
    # noinspection PyUnresolvedReferences
    from local_settings import *  # noqa

# =============================================================================
# Extra actions from the site-specific file
# =============================================================================

if CRATE_HTTPS:
    # https://docs.djangoproject.com/en/1.8/ref/settings/
    # We could do
    #   SECURE_SSL_REDIRECT = True  # redirect all HTTP to HTTPS
    #   SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
    #   ... trust the X-Forwarded-Proto header from the front-end browser;
    #       if it's 'https', we trust the connection is secure
    # but it is a bit tricky to get right:
    #   https://docs.djangoproject.com/en/1.8/ref/settings/#std:setting-SECURE_PROXY_SSL_HEADER  # noqa
    # Instead, YOU SHOULD RESTRICT THE FRONT END. See instructions.txt.
    SESSION_COOKIE_SECURE = True  # cookies only via HTTPS
    CSRF_COOKIE_SECURE = True  # CSRF cookies only via HTTPS



#!/usr/bin/env python
# crate_anon/nlp_manager/text_handling.py

from typing import Dict
from cardinal_pythonlib.rnc_lang import chunks


# =============================================================================
# Input support methods
# =============================================================================

def tsv_pairs_to_dict(line: str, key_lower: bool = True) -> Dict[str, str]:
    """
    Converts a TSV line into sequential key/value pairs as a dictionary.
    """
    items = line.split("\t")
    d = {}
    for chunk in chunks(items, 2):
        key = chunk[0]
        value = unescape_tabs_newlines(chunk[1])
        if key_lower:
            key = key.lower()
        d[key] = value
    return d


def escape_tabs_newlines(s: str) -> str:
    """
    Escapes CR, LF, tab, and backslashes. (Here just for testing; mirrors the
    equivalent function in the Java code.)
    """
    if not s:
        return s
    s = s.replace("\\", r"\\")  # replace \ with \\
    s = s.replace("\n", r"\n")  # escape \n; note ord("\n") == 10
    s = s.replace("\r", r"\r")  # escape \r; note ord("\r") == 13
    s = s.replace("\t", r"\t")  # escape \t; note ord("\t") == 9
    return s


def unescape_tabs_newlines(s: str) -> str:
    """
    Reverses escape_tabs_newlines.
    """
    # See also http://stackoverflow.com/questions/4020539
    if not s:
        return s
    d = ""  # the destination string
    in_escape = False
    for i in range(len(s)):
        c = s[i]  # the character being processed
        if in_escape:
            if c == "r":
                d += "\r"
            elif c == "n":
                d += "\n"
            elif c == "t":
                d += "\t"
            else:
                d += c
            in_escape = False
        else:
            if c == "\\":
                in_escape = True
            else:
                d += c
    return d



#!/usr/bin/env python
# crate_anon/crateweb/consent/__init__.py

# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.
from crate_anon.crateweb.consent.celery import app as celery_app  # noqa



#!/usr/bin/env python
# crate_anon/crateweb/consent/emails.py

from django.conf import settings
from crate_anon.crateweb.consent.models import (
    ContactRequest,
    ClinicianResponse,
)
from crate_anon.crateweb.consent.utils import render_email_html_to_string


def email_clinician_html(contact_request: ContactRequest) -> str:
    return render_email_html_to_string('email_clinician.html', {
        'ClinicianResponse': ClinicianResponse,
        'consent_mode': contact_request.consent_mode,
        'contact_request': contact_request,
        'patient_lookup': contact_request.patient_lookup,
        'settings': settings,
        'study': contact_request.study,

        # 'url_yes': XXX,
        # 'url_no': XXX,
        # 'url_maybe': XXX,
    })



#!/usr/bin/env python
# crate_anon/crateweb/config/urls.py

"""crateweb URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.8/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Add an import:  from blog import urls as blog_urls
    2. Add a URL to urlpatterns:  url(r'^blog/', include(blog_urls))
"""
from django.conf.urls import include, url  # patterns
# import django.contrib.auth.views
# from django.contrib import admin
# import django.views.defaults
# import admin_honeypot
from crate_anon.crateweb.core.admin import (
    mgr_admin_site,
    dev_admin_site,
    res_admin_site,
)
import crate_anon.crateweb.core.auth_views as core_auth_views
import crate_anon.crateweb.core.views as core_views
import crate_anon.crateweb.consent.views as consent_views
import crate_anon.crateweb.research.views as research_views
import crate_anon.crateweb.userprofile.views as userprofile_views

# This is the place for one-time startup code.
# http://stackoverflow.com/questions/6791911/execute-code-when-django-starts-once-only  # noqa
# So we cache things here that we don't want the user to have to wait for:
from crate_anon.crateweb.research.models import research_database_info

research_database_info.get_infodictlist()


urlpatterns = [
    # -------------------------------------------------------------------------
    # Login, other authentication/password stuff
    # -------------------------------------------------------------------------
    url(r'^login/', core_auth_views.login_view, name='login'),
    url(r'^logout/', core_auth_views.logout_view, name='logout'),
    url(r'^password_change/', core_auth_views.password_change,
        name='password_change'),

    # -------------------------------------------------------------------------
    # Home, About
    # -------------------------------------------------------------------------
    url(r'^$', core_views.home, name='home'),
    url(r'^about/$', core_views.about, name='about'),

    # -------------------------------------------------------------------------
    # Admin sites
    # -------------------------------------------------------------------------
    # ... obfuscate: p351 of Greenfeld_2015.
    url(r'^mgr_admin/', include(mgr_admin_site.urls)),
    url(r'^dev_admin/', include(dev_admin_site.urls)),
    url(r'^res_admin/', include(res_admin_site.urls)),
    # ... namespace is defined in call to AdminSite(); see core/admin.py

    # -------------------------------------------------------------------------
    # Main query views
    # -------------------------------------------------------------------------
    url(r'^build_query/$', research_views.build_query, name='build_query'),
    url(r'^query/$', research_views.edit_select_query, name='query'),
    url(r'^activate_query/(?P<query_id>[0-9]+)/$',
        research_views.activate_query, name='activate_query'),
    url(r'^delete_query/(?P<query_id>[0-9]+)/$',
        research_views.delete_query, name='delete_query'),
    url(r'^highlight/$',
        research_views.edit_select_highlight, name='highlight'),
    url(r'^activate_highlight/(?P<highlight_id>[0-9]+)/$',
        research_views.activate_highlight, name='activate_highlight'),
    url(r'^deactivate_highlight/(?P<highlight_id>[0-9]+)/$',
        research_views.deactivate_highlight, name='deactivate_highlight'),
    url(r'^delete_highlight/(?P<highlight_id>[0-9]+)/$',
        research_views.delete_highlight, name='delete_highlight'),
    url(r'^count/(?P<query_id>[0-9]+)/$',
        research_views.count, name='count'),
    url(r'^results/(?P<query_id>[0-9]+)/$',
        research_views.results, name='results'),
    url(r'^results_recordwise/(?P<query_id>[0-9]+)/$',
        research_views.results_recordwise, name='results_recordwise'),
    url(r'^tsv/$', research_views.tsv, name='tsv'),

    # -------------------------------------------------------------------------
    # Research database structure
    # -------------------------------------------------------------------------
    url(r'^structure_table_long/$', research_views.structure_table_long,
        name='structure_table_long'),
    url(r'^structure_table_paginated/$',
        research_views.structure_table_paginated,
        name='structure_table_paginated'),
    url(r'^structure_tree/$', research_views.structure_tree,
        name='structure_tree'),
    url(r'^structure_tsv/$', research_views.structure_tsv,
        name='structure_tsv'),

    # -------------------------------------------------------------------------
    # SQL helpers
    # -------------------------------------------------------------------------
    url(r'^sqlhelper_text_anywhere/$', research_views.sqlhelper_text_anywhere,
        name='sqlhelper_text_anywhere'),

    # -------------------------------------------------------------------------
    # Researcher consent functions
    # -------------------------------------------------------------------------
    url(r'^submit_contact_request/$', consent_views.submit_contact_request,
        name='submit_contact_request'),

    # -------------------------------------------------------------------------
    # Look up PID
    # -------------------------------------------------------------------------
    url(r'^pidlookup/$', research_views.pidlookup, name='pidlookup'),

    # -------------------------------------------------------------------------
    # User profile
    # -------------------------------------------------------------------------
    url(r'^edit_profile/$', userprofile_views.edit_profile,
        name='edit_profile'),

    # -------------------------------------------------------------------------
    # Superuser access only
    # -------------------------------------------------------------------------
    url(r'^download_privatestorage/(?P<filename>.+)$',
        consent_views.download_privatestorage,
        name="download_privatestorage"),
        # ... NB hard-coded reference to this in consent/storage.py;
        # can't use reverse
    url(r'^charity_report/$',
        consent_views.charity_report, name="charity_report"),
    url(r'^exclusion_report/$',
        consent_views.exclusion_report, name="exclusion_report"),
    url(r'^test_email_rdbm/$',
        consent_views.test_email_rdbm, name="test_email_rdbm"),

    # -------------------------------------------------------------------------
    # Public views
    # -------------------------------------------------------------------------
    url(r'^study_details/(?P<study_id>[0-9]+)/$', consent_views.study_details,
        name='study_details'),
    url(r'^study_form/(?P<study_id>[0-9]+)/$', consent_views.study_form,
        name='study_form'),
    url(r'^study_pack/(?P<study_id>[0-9]+)/$', consent_views.study_pack,
        name='study_pack'),
    url(r'^leaflet/(?P<leaflet_name>[a-zA-Z0-9_]+)/$',
        consent_views.view_leaflet, name='leaflet'),

    # -------------------------------------------------------------------------
    # Restricted views (token-based); clinicians
    # -------------------------------------------------------------------------
    url(r'^clinician_response/(?P<clinician_response_id>-?[0-9]+)/$',
        consent_views.clinician_response_view, name='clinician_response'),
        # note the -? : allows viewing (and URL-reversing within) an e-mail
        # having a dummy ID of -1.
    url(r'^clinician_pack/(?P<clinician_response_id>[0-9]+)/(?P<token>[a-zA-Z0-9]+)/$',  # noqa
        consent_views.clinician_pack, name='clinician_pack'),

    # -------------------------------------------------------------------------
    # Restricted views; superuser + researchers
    # -------------------------------------------------------------------------
    url(r'^view_email_html/(?P<email_id>[0-9]+)/$',
        consent_views.view_email_html, name='view_email_html'),
    url(r'^view_email_attachment/(?P<attachment_id>[0-9]+)/$',
        consent_views.view_email_attachment, name='view_email_attachment'),
    url(r'^letter/(?P<letter_id>[0-9]+)/$',
        consent_views.view_letter, name='letter'),

    # -------------------------------------------------------------------------
    # Developer functions and test views
    # -------------------------------------------------------------------------
    url(r'^generate_fake_nhs/$', consent_views.generate_fake_nhs,
        name='generate_fake_nhs'),
    url(r'^test_patient_lookup/$', consent_views.test_patient_lookup,
        name='test_patient_lookup'),

    url(r'^draft_clinician_email/(?P<contact_request_id>[0-9]+)/$',
        consent_views.draft_clinician_email,
        name='draft_clinician_email'),
    url(r'^draft_approval_email/(?P<contact_request_id>[0-9]+)/$',
        consent_views.draft_approval_email,
        name='draft_approval_email'),
    url(r'^draft_withdrawal_email/(?P<contact_request_id>[0-9]+)/$',
        consent_views.draft_withdrawal_email,
        name='draft_withdrawal_email'),

    url(r'^draft_approval_letter/(?P<contact_request_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.draft_approval_letter,
        name='draft_approval_letter'),
    url(r'^draft_withdrawal_letter/(?P<contact_request_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.draft_withdrawal_letter,
        name='draft_withdrawal_letter'),
    url(r'^draft_first_traffic_light_letter/(?P<patient_lookup_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.draft_first_traffic_light_letter,
        name='draft_first_traffic_light_letter'),
    url(r'^draft_letter_clinician_to_pt_re_study/(?P<contact_request_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.draft_letter_clinician_to_pt_re_study,
        name='draft_letter_clinician_to_pt_re_study'),
    url(r'^decision_form_to_pt_re_study/(?P<contact_request_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.decision_form_to_pt_re_study,
        name='decision_form_to_pt_re_study'),
    url(r'^draft_confirm_traffic_light_letter/(?P<consent_mode_id>[0-9]+)/(?P<viewtype>pdf|html)/$',  # noqa
        consent_views.draft_confirm_traffic_light_letter,
        name='draft_confirm_traffic_light_letter'),

    # -------------------------------------------------------------------------
    # Other test views
    # -------------------------------------------------------------------------
    # url(r'^404/$', django.views.defaults.page_not_found, ),

]

# urlpatterns += patterns('', url(r'^silk/',
#                                 include('silk.urls', namespace='silk')))



#!/usr/bin/env python
# setup.py

"""
CRATE setup file

To use:

    python setup.py sdist

    twine upload dist/*

To install in development mode:

    pip install -e .

"""
# https://packaging.python.org/en/latest/distributing/#working-in-development-mode  # noqa
# http://python-packaging-user-guide.readthedocs.org/en/latest/distributing/
# http://jtushman.github.io/blog/2013/06/17/sharing-code-across-applications-with-python/  # noqa

from setuptools import setup, find_packages
from codecs import open
# import fnmatch
import os
import platform

from crate_anon.version import VERSION

here = os.path.abspath(os.path.dirname(__file__))

# setup.py is executed on the destination system at install time, so:
windows = platform.system() == 'Windows'

# -----------------------------------------------------------------------------
# Get the long description from the README file
# -----------------------------------------------------------------------------
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# -----------------------------------------------------------------------------
# Get all filenames
# -----------------------------------------------------------------------------
# rootdir = os.path.join(here, 'crate')
# data_files = []
# for dir_, subdirs, filenames in os.walk(rootdir):
#     files = []
#     reldir = os.path.relpath(dir_, rootdir)
#     for pattern in ['*.py', '*.html']:
#         for filename in fnmatch.filter(filenames, pattern):
#             files.append(filename)
#     if files:
#         data_files.append((reldir, files))
# print(data_files)
# http://stackoverflow.com/questions/2186525/use-a-glob-to-find-files-recursively-in-python  # noqa
# http://stackoverflow.com/questions/27664504/how-to-add-package-data-recursively-in-python-setup-py  # noqa

# rootdir = os.path.join(here, 'crate', 'crateweb', 'static_collected')
# static_collected = []
# for dir_, subdirs, filenames in os.walk(rootdir):
#     reldir = os.path.normpath(os.path.join(
#         'static_collected', os.path.relpath(dir_, rootdir)))
#     for filename in filenames:
#         if filename in ['.gitignore']:
#             continue
#         static_collected.append(os.path.join(reldir, filename))

# -----------------------------------------------------------------------------
# setup args
# -----------------------------------------------------------------------------
setup(
    name='crate-anon',  # 'crate' is taken

    version=VERSION,

    description='CRATE: clinical records anonymisation and text extraction',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/RudolfCardinal/crate',

    # Author details
    author='Rudolf Cardinal',
    author_email='rudolf@pobox.com',

    # Choose your license
    license='Apache License 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        'Natural Language :: English',

        'Operating System :: OS Independent',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',

        'Topic :: System :: Hardware',
        'Topic :: System :: Networking',
    ],

    keywords='anonymisation',

    packages=find_packages(),
    package_data={
        '': [
            'README.md'
        ],
        'crate_anon.crateweb': [
            # Don't use 'static/*', or at the point of installation it gets
            # upset about "demo_logo" ("can't copy... doesn't exist or not
            # a regular file). Keep running "python setup.py sdist >/dev/null"
            # until stderr comes up clean.
            'consent/templates/*.html',
            'consent/templates/*.js',
            'research/templates/*.html',
            'static/*.gif',
            'static/*.ico',
            'static/*.png',
            'static/demo_logo/*',
            'templates/*.css',
            'templates/*.html',
            'templates/*.js',
            'templates/admin/*.html',
            'userprofile/templates/*.html',
        ],
        'crate_anon.nlp_manager': [
            '*.java',
            '*.sh',
        ],
        'crate_anon.mysql_auditor': [
            '*.conf',
            '*.lua',
            '*.sh',
        ],
    },

    install_requires=[

        # ---------------------------------------------------------------------
        # For the web front end:
        # ---------------------------------------------------------------------
        # Core tools with accessories:
        'arrow==0.7.0',  # better datetime
        'colorlog==2.6.1',  # colour in logs
        # 'configobj==5.0.6',  # better config files
        'distro==0.6.0',  # replaces platform.linux_distribution
        # requires VC++ under Windows # 'mmh3==2.2',  # MurmurHash, for fast non-cryptographic hashing  # noqa
        'psutil==4.1.0',  # process management
        'pyhashxx==0.1.3',  # fast non-cryptographic hashing
        'semver==2.4.1',  # comparing semantic versions
        'SQLAlchemy>=1.1.3',  # database access
        'typing==3.5.2.2',  # part of stdlib in Python 3.5, but not 3.4

        # Web app:
        'celery==3.1.23',
        'colorlog==2.6.1',
        'Django>=1.9.3',  # "django" versus "Django": neither helps pycharm checking  # noqa
        'django-debug-toolbar==1.4',
        'django-debug-toolbar-template-profiler==1.0.1',
        'django-extensions==1.5.9',
        'django-picklefield==0.3.2',
        # 'django-silk==0.5.7',
        'django-sslserver==0.15',
        'flower==0.9.1',  # debug Celery; web server; only runs explicitly
        'pdfkit==0.5.0',
        # 'pygraphviz==1.3.1',  # not used
        'pyparsing==2.1.5',  # generic grammar parser
        'PyPDF2==1.25.1',
        'pytz==2015.6',
        'python-dateutil==2.4.2',
        'sqlparse==0.1.19',
        'Werkzeug==0.10.4',

        # ONSPD:
        'openpyxl==2.3.5',
        'xlrd==1.0.0',

        # Serving:
        'gunicorn==19.3.0',  # UNIX only, though will install under Windows
        'cherrypy==5.1.0',  # Cross-platform

        # ---------------------------------------------------------------------
        # For the anonymiser/pythonlib:
        # ---------------------------------------------------------------------

        'cardinal_pythonlib==0.2.4',

        'beautifulsoup4==4.4.1',
        'prettytable==0.7.2',
        # 'python-docx==0.8.5',  # needs lxml, which has Visual C++ dependencies under Windows  # noqa
        # ... https://python-docx.readthedocs.org/en/latest/user/install.html
        'regex==2015.11.14',
        'sortedcontainers==1.4.2',

        # ---------------------------------------------------------------------
        # For database connections (see manual): install manually
        # ---------------------------------------------------------------------
        # MySQL: one of:
        #   'PyMySQL',
        #   'mysqlclient',
        # SQL Server / ODBC route:
        #   'django-pyodbc-azure',
        #   'pyodbc',  # has C prerequisites
        #   'pypyodbc==1.3.3',
        # SQL Server / Embedded FreeTDS route:
        #   'django-pymssql',
        #   'django-mssql',
        #   'pymssql',
        # PostgreSQL:
        #   'psycopg2',  # has prerequisites (e.g. pg_config executable)
    ] + ([
        'pypiwin32==219'
    ] if windows else []),

    entry_points={
        'console_scripts': [
            # Format is 'script=module:function".
            'crate_make_demo_database=crate_anon.anonymise.make_demo_database:main',  # noqa
            'crate_print_demo_anon_config=crate_anon.anonymise.print_demo_config:main',  # noqa
            'crate_anonymise=crate_anon.anonymise.anonymise_cli:main',
            'crate_anonymise_multiprocess=crate_anon.anonymise.launch_multiprocess_anonymiser:main',  # noqa
            'crate_test_anonymisation=crate_anon.anonymise.test_anonymisation:main',  # noqa
            'crate_nlp=crate_anon.nlp_manager.nlp_manager:main',
            'crate_nlp_multiprocess=crate_anon.nlp_manager.launch_multiprocess_nlp:main',  # noqa
            'crate_nlp_build_gate_java_interface=crate_anon.nlp_manager.build_gate_java_interface:main',  # noqa
            'crate_nlp_build_medex_java_interface=crate_anon.nlp_manager.build_medex_java_interface:main',  # noqa
            'crate_nlp_build_medex_itself=crate_anon.nlp_manager.build_medex_itself:main',  # noqa

            'crate_generate_new_django_secret_key=crate_anon.tools.generate_new_django_secret_key:main',  # noqa
            'crate_estimate_mysql_memory_usage=crate_anon.tools.estimate_mysql_memory_usage:main',  # noqa

            'crate_django_manage=crate_anon.crateweb.manage:main',  # will cope with argv  # noqa
            'crate_launch_django_server=crate_anon.crateweb.manage:runserver',
            'crate_launch_cherrypy_server=crate_anon.tools.launch_cherrypy_server:main',  # noqa
            # ... a separate script with ":main" rather than
            # "crate_anon.crateweb.manage:runcpserver" so that we can launch
            # the "runcpserver" function from our Windows service, and have it
            # deal with the CherryPy special environment variable
            'crate_launch_celery=crate_anon.tools.launch_celery:main',
            'crate_launch_flower=crate_anon.tools.launch_flower:main',
            'crate_print_demo_crateweb_config=crate_anon.tools.print_crateweb_demo_config:main',  # noqa

            'crate_windows_service=crate_anon.tools.winservice:main',

            'crate_postcodes=crate_anon.preprocess.postcodes:main',
            'crate_preprocess_rio=crate_anon.preprocess.preprocess_rio:main',
        ],
    },
)



#!/usr/bin/env python
# crate_anon/crateweb/research/sql_writer.py

import logging
from typing import Dict, List, Optional, Tuple

from pyparsing import ParseResults
import sqlparse

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger
from crate_anon.common.sql import (
    combine_db_table,
    split_db_table,
)
from crate_anon.common.sql_grammar_mysql import (
    column_spec,
    expr,
    # flatten,
    join_constraint,
    join_op,
    select_statement,
    table_spec,
    text_from_parsed,
)
from crate_anon.crateweb.research.models import (
    get_schema_trid_field,
    get_schema_rid_field,
    get_schema_rid_family,
    get_schema_mrid_table,
    get_schema_mrid_field,
)

log = logging.getLogger(__name__)


def parser_add_result_column(parsed: ParseResults,
                             column: str) -> ParseResults:
    # Presupposes at least one column already in the SELECT statement.

    existing_columns = parsed.select_expression.select_columns.asList()
    # log.critical(parsed.dump())
    # log.critical("existing columns: {}".format(repr(existing_columns)))
    # log.critical("adding column: {}".format(column))
    if column not in existing_columns:
        # log.critical("... doesn't exist; adding")
        newcol = column_spec.parseString(column)[0]
        parsed.select_expression.extend([",", newcol])
    # else:
    #     log.critical("... skipping column; exists")
    # log.critical(parsed.dump())
    return parsed


def parser_add_from_tables(parsed: ParseResults,
                           joininfo: List[Dict[str, str]]) -> ParseResults:
    """
    joininfo: list of dictionaries with keys:
        table, join_type, join_condition
    Presupposes at least one table already in the FROM clause.
    """
    # log.critical(parsed.dump())
    existing_tables = parsed.join_source.from_tables.asList()
    # log.critical("existing tables: {}".format(existing_tables))
    # log.critical("adding table: {}".format(table))
    for infodict in joininfo:
        table = infodict['table']
        join_type = infodict.get('join_type', 'INNER JOIN')
        join_condition = infodict.get('join_condition', None)
        if table in existing_tables:  # already there
            # log.critical("field already present")
            continue
        parsed_join = join_op.parseString(join_type)[0]  # e.g. INNER JOIN
        parsed_table = table_spec.parseString(table)[0]
        extrabits = [parsed_join, parsed_table]
        if join_condition:  # e.g. ON x = y
            extrabits.append(join_constraint.parseString(join_condition)[0])
        parsed.join_source.extend(extrabits)
    # log.critical(parsed.dump())
    return parsed


def get_first_from_db_table(
        parsed: ParseResults,
        match_db: bool = False,
        match_db_table: bool = False,
        db: str = None,
        table: str = None) -> Tuple[Optional[str], Optional[str]]:
    existing_tables = parsed.join_source.from_tables.asList()
    for t in existing_tables:
        db_component, table_component = split_db_table(t)
        if ((not match_db and not match_db_table) or
                (match_db_table and db_component == db and
                 table_component == table) or
                (match_db and db_component == db)):
            return db_component, table_component
    return None, None


def toggle_distinct(sql: str,
                    formatted: bool = True,
                    debug: bool = False,
                    debug_verbose: bool = False) -> str:
    p = select_statement.parseString(sql)
    if debug:
        log.info("START: {}".format(sql))
        if debug_verbose:
            log.debug("start dump:\n" + p.dump())
    # log.critical(repr(p.select_specifier))
    if p.select_specifier and 'DISTINCT' in p.select_specifier[0]:
        # log.critical("Already has DISTINCT")
        del p.select_specifier[:]
    else:
        # log.critical("Does not already have DISTINCT")
        p.select_specifier.append('DISTINCT')
    result = text_from_parsed(p, formatted=formatted)
    if debug:
        log.info("END: {}".format(result))
        if debug_verbose:
            log.debug("end dump:\n" + p.dump())
    return result


def get_join_info(parsed: ParseResults,
                  joindb: str,
                  jointable: str,
                  magic_join: bool = False,
                  nonmagic_join_type: str = "INNER JOIN",
                  nonmagic_join_condition: str = None) -> List[Dict[str, str]]:
    join_db_table = combine_db_table(joindb, jointable)
    first_db, first_table = get_first_from_db_table(parsed)
    db_match_db, db_match_table = get_first_from_db_table(
        parsed, match_db=True, db=joindb)
    _, table_match_table = get_first_from_db_table(
        parsed, match_db_table=True, db=joindb, table=jointable)

    if not first_table:
        # No tables in query yet.
        # This should not happen; this function is to help with adding
        # new FROM tables to existing FROM clauses.
        log.warning("get_join_info: no tables in query")
        return []
    if table_match_table:
        # This table is already in the query. No JOIN should be required.
        # log.critical("get_join_info: same table already in query")
        return []
    if not magic_join:
        # log.critical("get_join_info: non-magic join")
        return [{
            'join_type': nonmagic_join_type,
            'table': join_db_table,
            'join_condition': nonmagic_join_condition,
        }]
    if db_match_table:
        # Another table from the same database is present. Link on the
        # TRID field.
        # log.critical("get_join_info: joining to another table in same DB")
        return [{
            'join_type': 'INNER JOIN',
            'table': join_db_table,
            'join_condition': (
                "ON {new}.{trid} = {existing}.{trid}".format(
                    new=join_db_table,
                    existing=combine_db_table(db_match_db, db_match_table),
                    trid=get_schema_trid_field(db_match_db),
                )
            ),
        }]
    # OK. So now we're building a cross-database join.
    existing_family = get_schema_rid_family(first_db)
    new_family = get_schema_rid_family(joindb)
    # log.critical("existing_family={}, new_family={}".format(
    #     existing_family, new_family))
    if existing_family and existing_family == new_family:
        # log.critical("get_join_info: new DB, same RID family")
        return [{
            'join_type': 'INNER JOIN',
            'table': join_db_table,
            'join_condition': (
                "ON {new}.{newrid} = {existing}.{exrid}".format(
                    new=join_db_table,
                    newrid=get_schema_rid_field(joindb),
                    existing=combine_db_table(first_db, first_table),
                    exrid=get_schema_rid_field(first_db),
                )
            ),
        }]
    # If we get here, we have to do a complicated join via the MRID.
    # log.critical("get_join_info: new DB, different RID family, using MRID")
    existing_mrid_table = get_schema_mrid_table(first_db)
    existing_mrid_field = get_schema_mrid_field(first_db)
    new_mrid_table = get_schema_mrid_table(joindb)
    new_mrid_field = get_schema_mrid_field(joindb)

    existing_mrid_table_in_query = bool(get_first_from_db_table(
        parsed, match_db_table=True,
        db=first_db, table=existing_mrid_table)[1])

    joins = []
    if not existing_mrid_table_in_query:
        joins.append({
            'join_type': 'INNER JOIN',
            'table': combine_db_table(first_db, existing_mrid_table),
            'join_condition': (
                "ON {m1}.{trid1} = {t1}.{trid1}".format(
                    m1=combine_db_table(first_db, existing_mrid_table),
                    t1=combine_db_table(first_db, first_table),
                    trid1=get_schema_trid_field(first_db),
                )
            ),
        })
    joins.append({
        'join_type': 'INNER JOIN',
        'table': combine_db_table(joindb, new_mrid_table),
        'join_condition': (
            "ON {m2}.{mrid2} = {m1}.{mrid1}".format(
                m2=combine_db_table(joindb, new_mrid_table),
                mrid2=new_mrid_field,
                m1=combine_db_table(first_db, existing_mrid_table),
                mrid1=existing_mrid_field,
            )
        ),
    })
    if jointable != new_mrid_table:
        joins.append({
            'join_type': 'INNER JOIN',
            'table': join_db_table,
            'join_condition': (
                "ON {t2}.{trid2} = {m2}.{trid2}".format(
                    t2=combine_db_table(joindb, jointable),
                    m2=combine_db_table(joindb, new_mrid_table),
                    trid2=get_schema_trid_field(joindb),
                )
            ),
        })
    return joins


def add_to_select(sql: str,
                  # For SELECT:
                  select_db: str = None,
                  select_table: str = None,
                  select_column: str = None,
                  # For WHERE:
                  where_expression: str = None,
                  where_type: str = "AND",
                  where_db: str = None,
                  where_table: str = None,
                  bracket_where: bool = False,
                  # For either, for JOIN:
                  magic_join: bool = True,
                  join_type: str = "NATURAL JOIN",
                  join_condition: str = None,
                  # General:
                  formatted: bool = True,
                  debug: bool = False,
                  debug_verbose: bool = False) -> str:
    """
    This function encapsulates our query builder's common operations.
    One premise is that SQL parsing is relatively slow, so we should do this
    only once. We parse; add bits to the parsed structure as required; then
    re-convert to text.

    If you specify table/column, elements will be added to SELECT and FROM
    unless they already exist.

    If you specify where_expression, elements will be added to WHERE.
    In this situation, you should also specify where_table; if the where_table
    isn't yet in the FROM clause, this will be added as well.
    """

    if debug:
        log.info("START: {}".format(sql))
        log.debug("table: {}".format(select_table))
        log.debug("column: {}".format(select_column))
        log.debug("join_type: {}".format(join_type))
        log.debug("join_condition: {}".format(join_condition))
        log.debug("where_type: {}".format(where_type))
        log.debug("where_expression: {}".format(where_expression))
        log.debug("where_table: {}".format(where_table))
    if not sql:
        # ---------------------------------------------------------------------
        # Fresh SQL statement
        # ---------------------------------------------------------------------
        if debug:
            log.debug("Starting SQL from scratch")
        if select_column and select_table:
            db_table = combine_db_table(select_db, select_table)
            colspec = "{}.{}".format(db_table, select_column)
            result = sqlparse.format(
                "SELECT {} FROM {}".format(colspec, db_table),
                reindent=True)
            if where_expression:
                result += " WHERE {}".format(where_expression)
        else:
            raise ValueError("Blank starting SQL but no SELECT table/column")
    else:
        p = select_statement.parseString(sql)
        if debug and debug_verbose:
            log.debug("start dump:\n" + p.dump())

        # ---------------------------------------------------------------------
        # add SELECT... +/- FROM
        # ---------------------------------------------------------------------
        if select_column and select_table:
            colspec = "{}.{}".format(
                combine_db_table(select_db, select_table),
                select_column)
            p = parser_add_result_column(p, colspec)
            p = parser_add_from_tables(
                p, get_join_info(p, select_db, select_table,
                                 magic_join=magic_join))

        # ---------------------------------------------------------------------
        # add WHERE... +/- FROM
        # ---------------------------------------------------------------------
        if where_expression:
            cond = expr.parseString(where_expression)
            if p.where_clause:
                if bracket_where:
                    extra = [where_type, "(", cond, ")"]
                else:
                    extra = [where_type, cond]
                p.where_clause.where_expr.extend(extra)
            else:
                # No WHERE as yet
                if bracket_where:
                    extra = ["WHERE", "(", cond, ")"]
                else:
                    extra = ["WHERE", cond]
                p.where_clause.extend(extra)
            if where_table:
                p = parser_add_from_tables(
                    p, get_join_info(p, where_db, where_table,
                                     magic_join=magic_join,
                                     nonmagic_join_type=join_type,
                                     nonmagic_join_condition=join_condition))

        if debug and debug_verbose:
            log.debug("end dump:\n" + p.dump())
        result = text_from_parsed(p, formatted=formatted)
    if debug:
        log.info("END:\n{}".format(result))
    return result


def unit_tests() -> None:
    add_to_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
                  select_table="t2", select_column="c")
    add_to_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
                  select_table="t1", select_column="a")
    add_to_select("", select_table="t2", select_column="c")
    add_to_select("SELECT t1.a, t1.b FROM t1",
                  where_expression="t1.col2 < 3")
    add_to_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5",
                  where_expression="t1.col2 < 3")


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    unit_tests()



#!/usr/bin/env python
# crate_anon/common/debugfunc.py

import pdb
import sys
import traceback
from typing import Callable


def pdb_run(main: Callable[[], None]) -> None:
    # noinspection PyBroadException
    try:
        main()
    except:
        type_, value, tb = sys.exc_info()
        traceback.print_exc()
        pdb.post_mortem(tb)
